import logging
import tempfile
import uuid
import zipfile

import mantik.cli.utils as utils
import mantik.utils.mantik_api as mantik_api
import mantik.utils.urls

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def download_artifacts(
    project_id: uuid.UUID,
    run_id: uuid.UUID,
    target_dir: str,
    unzip: bool,
):
    artifacts_url = mantik_api.run.get_download_artifact_url(
        project_id=project_id,
        run_id=run_id,
        token=utils.access_token_from_env_vars(),
    )
    artifacts_path = mantik.utils.urls.get_local_path_from_url(
        target_dir=target_dir, url=artifacts_url, filetype=".zip"
    )
    if unzip:
        download_and_unzip_file_from_internet(
            artifacts_path=artifacts_path,
            artifacts_url=artifacts_url,
            target_dir=target_dir,
        )
    else:
        mantik.utils.urls.download_from_url(
            url=artifacts_url, target_path=artifacts_path
        )
        logger.info(f"Successfully downloaded at {artifacts_path}`")


def download_and_unzip_file_from_internet(
    artifacts_path: str,
    artifacts_url: str,
    target_dir: str,
):
    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_file = temp_dir_name + "/" + artifacts_path.split("/")[-1]
        mantik.utils.urls.download_from_url(
            url=artifacts_url, target_path=temp_file
        )
        with zipfile.ZipFile(temp_file, "r") as zip_ref:
            zip_ref.extractall(target_dir)
    logger.info(f"Artifacts successfully downloaded at {target_dir}`")



import pytest
import requests

import mantik.utils.mantik_api.experiment_repository as experiment_api


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_experiment(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    sample_experiment_repository_id,
    sample_experiment_repository,
    project_id,
):
    sample_schema = sample_experiment_repository.to_json()

    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/"
        f"experiments/{str(sample_experiment_repository_id)}",
        status_code=status_code,
        json_response=sample_schema,
        expected_error=expected,
    ) as (m, error):
        retrieved_model = experiment_api.get_one(
            project_id=project_id,
            experiment_repository_id=sample_experiment_repository_id,
            token="test_token",
        )
        assert retrieved_model == experiment_api.ExperimentRepository.from_json(
            sample_schema
        )
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


def test_get_unique_run_name(
    mock_mantik_api_request,
    info_caplog,
    sample_experiment_repository_id,
    project_id,
):
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/"
        f"experiments/{str(sample_experiment_repository_id)}/unique-mlflow-run-name",  # noqa
        status_code=200,
        json_response="unique-name",
    ) as (m, error):
        response = experiment_api.get_unique_run_name(
            project_id=project_id,
            experiment_repository_id=sample_experiment_repository_id,
            token="test_token",
            run_name="unique-name",
        )
        assert response == "unique-name"



import logging
import pathlib

import mantik.testing as testing
import mantik_compute_backend.firecrest as firecrest
import mantik_compute_backend.firecrest._exceptions as firecrest_exceptions


class TestFirecrestClient:
    def test_get_scratch_fails_for_not_scratch_or_home_given(
        self, expect_raise_if_exception
    ):
        machine = "test-machine"
        parameters = testing.firecrest.FakeParameters(
            storage=[
                testing.firecrest.FakeParameter(
                    name="FILESYSTEMS",
                    unit="",
                    value=[
                        {
                            "mounted": [
                                "/project",
                            ],
                            "system": machine,
                        },
                    ],
                    description="",
                ),
            ],
        )

        client = firecrest.client.Client(
            client=testing.firecrest.FakeClient(parameters=parameters),
            machine=machine,
            account=testing.firecrest.USER_NAME,
            sleep_time=0,
        )

        with expect_raise_if_exception(firecrest_exceptions.FirecrestError()):
            client._get_scratch_dir()

    def test_get_scratch_dir_for_duplicate_scratch_and_machine(self):
        machine = "test-machine"
        parameters = testing.firecrest.FakeParameters(
            storage=[
                testing.firecrest.FakeParameter(
                    name="FILESYSTEMS",
                    unit="",
                    value=[
                        {
                            "mounted": [
                                "/scratch",
                                "/scratch",
                            ],
                            "system": machine,
                        },
                        {
                            "mounted": [
                                "/scratch",
                                "/scratch",
                            ],
                            "system": machine,
                        },
                    ],
                    description="",
                ),
            ],
        )
        expected = pathlib.Path(
            f"/scratch/{testing.firecrest.USER_NAME}/mantik"
        )

        client = firecrest.client.Client(
            client=testing.firecrest.FakeClient(parameters=parameters),
            machine=machine,
            account=testing.firecrest.USER_NAME,
            sleep_time=0,
        )

        result = client._get_scratch_dir()

        assert result == expected

    def test_get_scratch_dir_from_home_if_no_scratch_given(self):
        machine = "test-machine"
        parameters = testing.firecrest.FakeParameters(
            storage=[
                testing.firecrest.FakeParameter(
                    name="FILESYSTEMS",
                    unit="",
                    value=[
                        {
                            "mounted": [
                                "/home",
                            ],
                            "system": machine,
                        },
                    ],
                    description="",
                ),
            ],
        )
        expected = pathlib.Path(f"/home/{testing.firecrest.USER_NAME}/mantik")

        client = firecrest.client.Client(
            client=testing.firecrest.FakeClient(parameters=parameters),
            machine=machine,
            account=testing.firecrest.USER_NAME,
            sleep_time=0,
        )

        result = client._get_scratch_dir()

        assert result == expected

    def test_upload_files_to_run_directory(self, caplog):
        caplog.set_level(logging.INFO)
        run_dir = pathlib.Path("/path/to/remote/mantik/mlflow-run-id")
        files = [
            pathlib.Path("/absolute/path/to/file1.py"),
            pathlib.Path("/absolute/path/to/folder/file2.py"),
        ]
        expected_source_and_target_paths = [
            "Uploading file /absolute/path/to/file1.py to "
            "/path/to/remote/mantik/mlflow-run-id",
            "Uploading file /absolute/path/to/folder/file2.py to "
            "/path/to/remote/mantik/mlflow-run-id/folder",
        ]
        client = firecrest.client.Client(
            client=testing.firecrest.FakeClient(),
            machine="daint",
            account="test-account",
            sleep_time=0,
        )

        client._upload_files_to_run_directory(run_dir=run_dir, files=files)

        for path in expected_source_and_target_paths:
            assert (
                path in caplog.text
            ), f"Path '{path}' not found in caplog.text"

    def test_upload_files_to_run_directory_fails(
        self, expect_raise_if_exception
    ):
        run_dir = pathlib.Path("/path/to/remote/mantik/mlflow-run-id")
        files = [
            pathlib.Path("/absolute/path/to/file1.py"),
        ]
        client = firecrest.client.Client(
            client=testing.firecrest.FakeClient(
                upload_status=testing.firecrest.UploadStatus.FAILED
            ),
            machine="daint",
            account="test-account",
            sleep_time=0,
        )

        with expect_raise_if_exception(firecrest_exceptions.FirecrestError()):
            client._upload_files_to_run_directory(run_dir=run_dir, files=files)



import dataclasses
import typing as t

import requests


def fake_get(
    monkeypatch,
    response_urls: t.Sequence[str],
    status_code: int = 200,
    expect_in_request_url: t.Optional[str] = None,
) -> None:
    """Fake `requests.get`."""
    urls_iter = iter(response_urls)

    def fake_response(url: str, *args, **kwargs):
        if (
            expect_in_request_url is not None
            and expect_in_request_url not in url
        ):
            raise ValueError(
                f"Expected '{expect_in_request_url}' to be present in the "
                f"request URL {url}"
            )

        return FakeResponse(url=next(urls_iter), status_code=status_code)

    monkeypatch.setattr(requests, "get", fake_response)


@dataclasses.dataclass
class FakeResponse(requests.Response):
    """Represents a response from the `requests` library."""

    url: str
    status_code: int
    reason = "test-reason"



import dataclasses
import typing as t

import mantik.utils.env as env

_MANTIK_USERNAME_ENV_VAR = "MANTIK_USERNAME"
_MANTIK_PASSWORD_ENV_VAR = "MANTIK_PASSWORD"


@dataclasses.dataclass(repr=False, frozen=True)
class Credentials:
    """Holds the credentials for authentication."""

    username: str
    password: str

    @classmethod
    def from_env(cls) -> "Credentials":
        """Create from environment variables."""
        username = env.get_required_env_var(_MANTIK_USERNAME_ENV_VAR)
        password = env.get_required_env_var(_MANTIK_PASSWORD_ENV_VAR)
        return cls(
            username=username,
            password=password,
        )

    def __str__(self) -> str:
        """Hide the credentials."""
        return self.__repr__()

    def __repr__(self) -> str:
        """Hide the credentials."""
        return f"{self.__class__.__name__}(<masked>)"

    def to_dict(self, include_password: bool = True) -> t.Dict:
        """Return as dict.

        Parameters
        ----------
        include_password : bool, default=True
            Whether to include the password in the result.

        """
        d = dataclasses.asdict(self)

        if include_password:
            return d

        d.pop("password")
        return d



PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"
DATA_REPOSITORY_ID_ENV_VAR = "MANTIK_DATA_REPOSITORY_ID"
DATA_REPOSITORY_VERSION_ENV_VAR = "MANTIK_DATA_REPOSITORY_VERSION"
DATA_REPOSITORY_COMMIT_ENV_VAR = "MANTIK_DATA_REPOSITORY_COMMIT"
DATA_REPOSITORY_BRANCH_ENV_VAR = "MANTIK_DATA_REPOSITORY_BRANCH"
EXPERIMENT_REPOSITORY_ID_ENV_VAR = "MANTIK_EXPERIMENT_REPOSITORY_ID"
CODE_REPOSITORY_ID_ENV_VAR = "MANTIK_CODE_REPOSITORY_ID"
MODEL_ID_ENV_VAR = "MANTIK_MODEL_ID"
CONNECTION_ID_ENV_VAR = "MANTIK_CONNECTION_ID"
RUN_ID_ENV_VAR = "MANTIK_RUN_ID"
MLFLOW_RUN_ID_ENV_VAR = "MLFLOW_RUN_ID"
TARGET_DIR_ENV_VAR = "MANTIK_TARGET_DIR"
MANTIK_ACCESS_TOKEN_ENV_VAR = "MLFLOW_TRACKING_TOKEN"



import os
import pathlib
import unittest.mock
import uuid

import click.testing
import pytest

import mantik.cli.data
import mantik.cli.main as main
import mantik.data_repository.data_repository


GROUP_NAME = mantik.cli.data.data.GROUP_NAME


@pytest.fixture
def cli_test_runner():
    yield click.testing.CliRunner()


@pytest.fixture(scope="function")
def sample_project_uuid():
    return uuid.uuid4()


@pytest.fixture(scope="function")
def mocked_token(mock_authentication):
    return mock_authentication


@pytest.fixture(scope="function")
def sample_data_repository_uuid():
    return uuid.uuid4()


@pytest.fixture(scope="function")
def sample_target_dir() -> str:
    return "/sample/target/dir"


@pytest.fixture(scope="function")
def sample_version() -> str:
    return "1234"


@pytest.fixture
def mock_project_id_env_var(sample_uuid_str):
    os.environ[mantik.utils.env_vars.PROJECT_ID_ENV_VAR] = sample_uuid_str
    yield
    os.unsetenv(mantik.utils.env_vars.PROJECT_ID_ENV_VAR)


def test_download_data_repository(
    mocked_token,
    cli_test_runner,
    sample_project_uuid,
    sample_data_repository_uuid,
    sample_target_dir,
    sample_version,
):
    with unittest.mock.patch(
        "mantik.data_repository.data_repository.download_data_repository"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "download",
                "--project-id",
                str(sample_project_uuid),
                "--data-repository-id",
                str(sample_data_repository_uuid),
                "--branch",
                sample_version,
                "--target-dir",
                sample_target_dir,
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called_once_with(
            project_id=sample_project_uuid,
            data_repository_id=sample_data_repository_uuid,
            checkout=sample_version,
            target_dir=pathlib.Path(sample_target_dir),
            token=mocked_token,
        )



import pathlib

import pytest

import mantik.config.environment as _environment
import mantik.config.executable as executable
import mantik.utils.unicore.upload as upload_files


@pytest.mark.parametrize(
    ("environment", "expect_image_is_included"),
    [
        # Test case: No environment given
        (
            None,
            False,
        ),
        # Test case: No execution environment given
        (
            _environment.Environment(),
            False,
        ),
        # Test case: upload local image
        (
            _environment.Environment(
                execution=executable.Apptainer(
                    path=pathlib.Path("mantik-test.sif"),
                    type="local",
                ),
            ),
            True,
        ),
        # Test case: use remote image, don't upload a local image
        (
            _environment.Environment(
                execution=executable.Apptainer(
                    path=pathlib.Path("/absolute/path/to/remote/image.sif"),
                    type="remote",
                ),
            ),
            False,
        ),
    ],
)
def test_get_files_to_upload(
    example_project_path,
    example_unicore_config,
    environment,
    expect_image_is_included,
):
    example_unicore_config.environment = environment
    files = upload_files.get_files_to_upload(
        project_dir=example_project_path, config=example_unicore_config
    )
    if expect_image_is_included:
        assert example_project_path / "mantik-test.sif" in files
    else:
        assert example_project_path / "mantik-test.sif" not in files


@pytest.mark.parametrize(
    ("exclude", "expected"),
    [
        # Test case: exclude all files
        (["*"], []),
        # Test case: exclude all Python files
        (
            [
                # pycache folder might be added when executing any file
                # in the project dir manually.
                "__pycache__/",
                "*.py",
            ],
            [
                "Dockerfile",
                "MLproject",
                "recipe.def",
                "compute-backend-config.json",
                "compute-backend-config.yaml",
                "compute-backend-firecrest-config.json",
                "compute-backend-firecrest-config.yaml",
                "mantik-test.sif",
                "config-with-errors.yaml",
            ],
        ),
        # Test case: exclude all except main.py and test_subfolder/test.py
        (
            [
                "__pycache__/",
                "Dockerfile",
                "MLproject",
                "recipe.def",
                "compute-backend-config.json",
                "compute-backend-config.yaml",
                "compute-backend-firecrest-config.json",
                "compute-backend-firecrest-config.yaml",
                "mantik-test.sif",
                "config-with-errors.yaml",
            ],
            ["main.py", "run.py", "test_subfolder/test.py"],
        ),
        # Test case: exclude all except main.py
        (
            [
                "__pycache__/",
                "Dockerfile",
                "MLproject",
                "recipe.def",
                "compute-backend-config.json",
                "compute-backend-config.yaml",
                "compute-backend-firecrest-config.json",
                "compute-backend-firecrest-config.yaml",
                "mantik-test.sif",
                "test_subfolder/",
                "config-with-errors.yaml",
            ],
            ["main.py", "run.py"],
        ),
        # Test case: exclude all except test_subfolder/test.py
        (
            [
                "__pycache__/",
                "Dockerfile",
                "MLproject",
                "recipe.def",
                "compute-backend-config.json",
                "compute-backend-config.yaml",
                "compute-backend-firecrest-config.json",
                "compute-backend-firecrest-config.yaml",
                "mantik-test.sif",
                "config-with-errors.yaml",
                "main.py",
                "run.py",
            ],
            ["test_subfolder/test.py"],
        ),
    ],
)
def test_recursively_list_files_in_directory(
    example_project_path, exclude, expected
):
    for i, element in enumerate(expected):
        expected[i] = example_project_path / element
    result = upload_files._recursively_list_files_in_directory(
        example_project_path, exclude
    )

    assert sorted(result) == sorted(expected)


@pytest.mark.parametrize(
    ("file_name", "exclude", "expected"),
    [
        ("main.py", ["*.py"], True),
        ("main.py", ["*.yaml"], False),
    ],
)
def test_file_matches_exclude_entry(
    file_name, example_project_path, exclude, expected
):
    filepath = example_project_path / file_name
    assert expected == upload_files._file_matches_exclude_entry(
        filepath, exclude
    )


@pytest.mark.parametrize(
    ("pattern", "expected"),
    [
        ("*.py", ["run.py", "main.py"]),
        ("**/*.py", ["run.py", "main.py", "test.py"]),
    ],
)
def test_files_matching_pattern(example_project_path, pattern, expected):
    result = upload_files._files_matching_pattern(example_project_path, pattern)

    assert sorted(result) == sorted(expected)


@pytest.mark.parametrize(
    ("element_relative_path", "exclude", "expected"),
    [
        (pathlib.Path("subfolder/test.py"), ["subfolder/"], True),
    ],
)
def test_file_in_excluded_subdirectory(
    element_relative_path, exclude, expected
):
    assert expected == upload_files._file_in_excluded_subdirectory(
        element_relative_path, exclude
    )



import click.testing
import pytest

import mantik
import mantik.cli.main as main
import mantik.testing.token as testing_token


@pytest.mark.parametrize(
    ("cli_args", "expected"),
    [
        ([], "export MLFLOW_TRACKING_TOKEN=test-access-token\n"),
        (["--no-export"], "MLFLOW_TRACKING_TOKEN=test-access-token\n"),
    ],
)
@testing_token.set_token()
def test_init_from_env_vars(tmp_dir_as_test_mantik_folder, cli_args, expected):
    runner = click.testing.CliRunner()
    result = runner.invoke(main.cli, ["init", *cli_args])

    assert result.exit_code == 0
    assert result.output == expected
    mantik.testing.env.assert_conflicting_mlflow_env_vars_not_set()
    mantik.testing.env.assert_correct_tracking_token_env_var_set(
        "test-access-token"
    )



import contextlib
import functools
import logging
import pathlib
import typing as t

import mantik.tracking.environment as _environment
import mantik.tracking.track as track

try:
    import mlflow
    import mlflow.entities
    import mlflow.entities.model_registry
    import mlflow.utils.file_utils
    import mlflow.projects.utils
except ModuleNotFoundError as exc:
    raise RuntimeError(
        "Install supported MLflow version (`pip install mantik[mlflow]`)"
        "or any desired MLflow version"
    ) from exc


logger = logging.getLogger(__name__)


def _catch_exceptions(func: t.Callable) -> t.Callable:
    """Catch MLflow exceptions and log them instead of raising."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> t.Any:
        try:
            return func(*args, **kwargs)
        except mlflow.exceptions.MlflowException:
            logger.warning(
                "MLflow API request failed for %s with args %s and kwargs %s",
                func.__name__,
                args,
                kwargs,
                exc_info=True,
            )
            return None

    return wrapper


def init_tracking() -> _environment.Environment:
    """Attempts to initialize tracking to Mantik, emits warning if it fails.

    If successful, this sets the `MLFLOW_TRACKING_TOKEN` environment variable
    to the access token retrieved from the Mantik API.

    Returns
    -------
    mantik.tracking.environment.Environment
        Holds the access token for the Mantik and MLflow API.

    """
    try:
        return track.init_tracking()
    except Exception:
        logger.warning(
            "Unable to retrieve access token from Mantik API", exc_info=True
        )


@contextlib.contextmanager
def start_run(
    run_id: str = None,
    experiment_id: t.Optional[str] = None,
    run_name: t.Optional[str] = None,
    nested: bool = False,
    tags: t.Optional[t.Dict[str, t.Any]] = None,
    description: t.Optional[str] = None,
) -> mlflow.ActiveRun:
    """Start a new MLflow run or pick up the run specified as `MLFLOW_RUN_ID`,
    which is set by the Mantik Compute Backend when submitting a run."""
    try:
        yield mlflow.start_run(
            run_id=run_id,
            experiment_id=experiment_id,
            run_name=run_name,
            nested=nested,
            tags=tags,
            description=description,
        )
    except mlflow.exceptions.MlflowException:
        logger.warning("Start MLflow run failed", exc_info=True)
        yield None


@_catch_exceptions
def active_run() -> t.Optional[mlflow.ActiveRun]:
    """Get the currently active run, or `None` if no such run exists."""
    return mlflow.active_run()


@_catch_exceptions
def end_run(
    status: str = mlflow.entities.RunStatus.to_string(
        mlflow.entities.RunStatus.FINISHED
    ),
) -> None:
    """End an active MLflow run."""
    return mlflow.end_run(status=status)


def delete_run(run_id: str) -> None:
    """Not supported: only permitted via the Mantik platform."""
    raise NotImplementedError(
        "Deleting a run is only permitted via the Mantik platform"
    )


def create_experiment(
    name: str,
    artifact_location: t.Optional[str] = None,
    tags: t.Optional[t.Dict[str, t.Any]] = None,
) -> None:
    """Not supported: only permitted via the Mantik platform."""
    raise NotImplementedError(
        "Creating an experiment is only permitted via the Mantik platform"
    )


def delete_experiment(experiment_id: int) -> None:
    """Not supported: only permitted via the Mantik platform."""
    raise NotImplementedError(
        "Deleting an experiment is only permitted via the Mantik platform"
    )


@_catch_exceptions
def log_artifact(
    local_path: str, artifact_path: t.Optional[str] = None
) -> None:
    """Log a local file or directory as an artifact of the currently
    active run."""
    return mlflow.log_artifact(
        local_path=local_path, artifact_path=artifact_path
    )


@_catch_exceptions
def log_artifacts(
    local_dir: str, artifact_path: t.Optional[str] = None
) -> None:
    """Log all the contents of a local directory as artifacts of the run."""
    return mlflow.log_artifacts(
        local_dir=local_dir, artifact_path=artifact_path
    )


@_catch_exceptions
def log_dict(dictionary: t.Any, artifact_file: str) -> t.Any:
    """Log a JSON/YAML-serializable object (e.g. dict) as an artifact."""
    return mlflow.log_dict(dictionary=dictionary, artifact_file=artifact_file)


@_catch_exceptions
def log_figure(figure, artifact_file: str) -> t.Any:
    """Log a figure as an artifact."""
    return mlflow.log_figure(figure=figure, artifact_file=artifact_file)


@_catch_exceptions
def log_image(image, artifact_file: str) -> t.Any:
    """Log an image as an artifact."""
    return mlflow.log_image(image=image, artifact_file=artifact_file)


@_catch_exceptions
def log_metric(key: str, value: float, step: t.Optional[int] = None) -> None:
    """Log a metric for the current run."""
    return mlflow.log_metric(key=key, value=value, step=step)


@_catch_exceptions
def log_metrics(
    metrics: t.Dict[str, float], step: t.Optional[int] = None
) -> None:
    """Log a batch of metrics for the current run."""
    return mlflow.log_metrics(metrics=metrics, step=step)


@_catch_exceptions
def log_param(key: str, value: t.Any) -> t.Any:
    """Log a parameter (e.g. model hyperparameter) for the current run."""
    return mlflow.log_param(key=key, value=value)


@_catch_exceptions
def log_params(params: t.Dict[str, t.Any]) -> None:
    """Log a batch of params for the current run."""
    return mlflow.log_params(params)


@_catch_exceptions
def log_text(text: str, artifact_file: str) -> None:
    """Log text as an artifact."""
    return mlflow.log_text(text=text, artifact_file=artifact_file)


@_catch_exceptions
def register_model(
    model_uri,
    name,
    *,
    tags: t.Optional[t.Dict[str, t.Any]] = None,
) -> mlflow.entities.model_registry.ModelVersion:
    """Create a new model version in the model registry."""
    return mlflow.register_model(model_uri=model_uri, name=name, tags=tags)


@_catch_exceptions
def set_tag(key: str, value: t.Any) -> None:
    """Log a tag for the current run."""
    return mlflow.set_tag(key=key, value=value)


@_catch_exceptions
def set_tags(tags: t.Dict[str, t.Any]) -> None:
    """Log a batch of tags for the current run."""
    return mlflow.set_tags(tags=tags)


@_catch_exceptions
def delete_tag(key: str) -> None:
    """Delete a tag."""
    return mlflow.delete_tag(key=key)


@_catch_exceptions
def set_tracking_uri(uri: t.Union[str, pathlib.Path]) -> None:
    """Set the tracking server URI (`MLFLOW_TRACKING_URI` environment
    variable).

    """
    return mlflow.set_tracking_uri(uri=uri)


@_catch_exceptions
def get_tracking_uri() -> str:
    """Get the MLflow tracking URI set as `MLFLOW_TRACKING_URI`."""
    return mlflow.get_tracking_uri()


@_catch_exceptions
def get_local_backend_config_run_id_env_var() -> str:
    """Get the MLflow local backend config run id environment variable."""
    return mlflow.projects.utils.MLFLOW_LOCAL_BACKEND_RUN_ID_CONFIG


def call_method(method: t.Union[t.Callable, str], *args, **kwargs) -> t.Any:
    """A general function to call any given MLflow method.

    Parameters
    ----------
    method : Callable or str
        MLflow method to invoke.

        If `str`, only the method name is requried, e.g.
        for `mlflow.log_param`, `method="log_param"` is required.

    args
        Passed to the MLflow method.
    kwargs
        Passed to the MLflow method.

    Examples
    --------
    With the mlflow method:

    >>> import mantik.mlflow
    >>> import mlflow
    >>> mantik.mlflow.call_method(mlflow.log_param, key="<name>", value="<value>")  # noqa: E501

    or with a `str`:

    >>> import mantik.mlflow as mlflow
    >>> mlflow.call_method("log_param", key="<name>", value="<value>")  # noqa: E501

    Returns
    -------
    The returned instance of the MLflow method.

    """
    if isinstance(method, str):
        method = getattr(mlflow, method)

    try:
        return method(*args, **kwargs)
    except mlflow.exceptions.MlflowException:
        logger.exception(
            "Calling mlflow.%s with args=%s and kwargs=%s has failed",
            method.__name__,
            args,
            kwargs,
            exc_info=True,
        )
        return None



import uuid

import mantik_compute_backend.models as models


def test_submit_run_response():
    run_id = uuid.uuid4()
    response = models.SubmitRunResponse(
        experiment_id=1,
        run_id=run_id,
        unicore_job_id="3",
        tracking_uri="foo.bar",
    )
    assert response.experiment_id == 1
    assert response.run_id == run_id
    assert response.unicore_job_id == "3"



import contextlib

import pytest

import mantik.utils.env as env
import mantik.utils.mantik_api.credentials as _credentials

# Variables have to be renamed for testing. Otherwise, the actual
# variables will be replaced and reset if it is set at the time of the test.
# This does not allow testing cases where the environment variables are not set
# by the user.
_credentials._MANTIK_USERNAME_ENV_VAR = "TEST_MANTIK_USERNAME_ENV_VAR"
_credentials._MANTIK_PASSWORD_ENV_VAR = "TEST_MANTIK_PASSWORD_ENV_VAR"
_credentials._MANTIK_SECRET_ENV_VAR = "TEST_MANTIK_SECRET_ENV_VAR"


class TestCredentials:
    @pytest.mark.parametrize(
        ("env_vars", "expected"),
        [
            (
                {
                    _credentials._MANTIK_USERNAME_ENV_VAR: "test-username",
                    _credentials._MANTIK_PASSWORD_ENV_VAR: "test-password",
                },
                _credentials.Credentials(
                    username="test-username",
                    password="test-password",
                ),
            ),
            # Test case: username env var not set
            (
                {
                    _credentials._MANTIK_USERNAME_ENV_VAR: "test-username",
                },
                NameError(),
            ),
            # Test case: password env var not set
            (
                {
                    _credentials._MANTIK_PASSWORD_ENV_VAR: "test-password",
                },
                NameError(),
            ),
        ],
    )
    def test_from_env(self, env_vars, expected):
        with env.env_vars_set(env_vars):
            with pytest.raises(type(expected)) if isinstance(
                expected, Exception
            ) else contextlib.nullcontext():
                result = _credentials.Credentials.from_env()
                assert result == expected



import contextlib
import logging
import os
import typing as t

logger = logging.getLogger(__name__)


def get_required_env_var(name: str) -> str:
    """Return a required environment variable.

    Parameter
    ---------
    name : str
        Name of the environment variable.

    Raises
    ------
    NameError
        If the environment variable is unset.

    Returns
    -------
    str
        The value of the environment variable.


    """
    value = os.getenv(name)
    if value is None:
        raise NameError(f"Environment variable '{name}' not set")
    return value


def get_optional_env_var(
    name: str, default: t.Optional[str] = None
) -> t.Optional[str]:
    """Return an optional environment variable.

    Parameter
    ---------
    name : str
        Name of the environment variable.
    default : str, optional
        The default to return

    Returns
    -------
    str or None
        The value of the environment variable.


    """
    value = os.getenv(name, default=default)
    if value is None:
        logger.debug("Optional environment variable %s not set", name)
    return value


@contextlib.contextmanager
def env_vars_set(env_vars: t.Dict[str, t.Any]) -> None:
    """Set the given environment variables and unset afterwards.

    Parameters
    ----------
    env_vars : dict
        Environment variables and values to set.

    Notes
    -----
    All environment variables that were previously set to another value will
    be reset to the initial value afterwards.

    """
    set_env_vars(env_vars)
    yield
    unset_env_vars(env_vars.keys())


def set_env_vars(env_vars: t.Dict[str, t.Any]) -> None:
    """Set given environment variables.

    Parameters
    ----------
    env_vars : dict
        Environment variables and values to set.

    """
    logger.debug("Setting environment variables %s", env_vars)
    for key, value in env_vars.items():
        if value is None:
            try:
                os.environ.pop(key)
            except KeyError:
                pass
        else:
            os.environ[key] = value


def unset_env_vars(env_vars: t.Iterable[str]) -> None:
    """Unset given environment variables.

    Parameters
    ----------
    env_vars : iterable
        Environment variables to be unset.

    """
    logger.debug("Unsetting environment variables %s", env_vars)
    for key in env_vars:
        try:
            os.environ.pop(key)
        except KeyError:
            # KeyError is raised if variable is already unset.
            pass


@contextlib.contextmanager
def env_vars_overwrite_temporarily(env_vars: t.Dict[str, t.Any]) -> None:
    original_env = {key: os.getenv(key) for key in env_vars.keys()}
    set_env_vars(env_vars)
    yield
    set_env_vars(original_env)



import abc
import functools
import typing as t


class ConfigObject(abc.ABC):
    """An object contained in the config."""

    @classmethod
    def from_dict(cls, config: t.Dict) -> "ConfigObject":
        """Initialize from dict."""
        return cls._from_dict(config)

    def to_dict(self) -> t.Dict:
        """Return as dict."""
        return _exclude_none_values(self._to_dict())

    @classmethod
    @abc.abstractmethod
    def _from_dict(cls, config: t.Dict) -> "ConfigObject":
        ...

    @abc.abstractmethod
    def _to_dict(self) -> t.Dict:
        ...


def _exclude_none_values(dict_) -> t.Dict:
    """Create a dict which only contains value that are not `None`."""
    return {
        key: _to_dict_if_config_object(value)
        for key, value in dict_.items()
        if value is not None
    }


@functools.singledispatch
def _to_dict_if_config_object(value: t.Any) -> t.Any:
    return value


@_to_dict_if_config_object.register
def _(value: ConfigObject) -> t.Any:
    return value.to_dict()



import uuid

import pytest
import requests

import mantik.utils.mantik_api as mantik_api
import mantik.utils.mantik_api.connection
import mantik.utils.mantik_api.data_repository
import mantik.utils.mlflow as mlflow


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_data_repositories(
    mock_mantik_api_request, info_caplog, status_code, expected
):
    project_id = uuid.uuid4()
    name = "data_repo_name"
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/data",
        status_code=status_code,
        json_response={
            "totalRecords": 1,
            "dataRepositories": [
                {
                    "dataRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                    "dataRepositoryName": name,
                    "uri": "string",
                    "accessToken": "string",
                    "description": "string",
                    "labels": [
                        {
                            "labelId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                            "scope": "string",
                            "name": "string",
                            "value": "string",
                        }
                    ],
                }
            ],
        },
        expected_error=expected,
    ) as (m, error):
        data_repositories = mantik.utils.mantik_api.data_repository.get_all(
            project_id=project_id, token="test_token"
        )
        assert data_repositories[0]["dataRepositoryName"] == name
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_connection(
    mock_mantik_api_request, info_caplog, status_code, expected
):
    user_id = uuid.uuid4()
    connection_id = uuid.uuid4()
    name = "unicore_connection"
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/users/{user_id}/settings/connections/{connection_id}",  # noqa
        status_code=status_code,
        json_response={
            "connectionId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "user": {
                "userId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                "name": "test-name",
                "email": "email-test",
            },
            "connectionName": name,
            "connectionProvider": "string",
            "authMethod": "string",
            "loginName": "string",
            "password": "string",
            "token": "string",
        },
        expected_error=expected,
    ) as (m, error):
        connection = mantik.utils.mantik_api.connection.get(
            user_id=user_id, connection_id=connection_id, token="test_token"
        )
        assert connection.connection_name == name
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("env_vars", "expected"),
    [
        (
            {
                mantik_api.client._MANTIK_API_URL_ENV_VAR: "https://api.cloud.mantik.ai"  # noqa F401
            },
            "https://api.cloud.mantik.ai",
        ),
        (
            {mlflow.TRACKING_URI_ENV_VAR: "https://tracking.cloud.mantik.ai"},
            "https://api.cloud.mantik.ai",
        ),
        ({}, mantik_api.client._DEFAULT_MANTIK_API_URL),
    ],
)
def test_get_base_url(env_vars_set, env_vars, expected):
    with env_vars_set(env_vars):
        result = mantik_api.client._get_base_url()

    assert result == expected



import uuid

import click

import mantik.cli._callbacks as _callbacks
import mantik.utils.env_vars as env_vars


VERBOSE = click.option(
    "--verbose",
    "-v",
    is_flag=True,
    callback=_callbacks.set_verbose_logging,
    help="Set logging to verbose mode.",
)

PROJECT_ID = click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=env_vars.PROJECT_ID_ENV_VAR,
    help=f"""Project ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.PROJECT_ID_ENV_VAR}.
    """,
)

MODEL_ID = click.option(
    "--model-id",
    type=uuid.UUID,
    required=True,
    envvar=env_vars.MODEL_ID_ENV_VAR,
    help=f"""Model ID on Mantik.

        If not specified, it is inferred from the environment variable
{env_vars.MODEL_ID_ENV_VAR}.""",
)


def get_connection_id(required: bool):
    return click.option(
        "--connection-id",
        required=required,
        default=None,
        type=uuid.UUID,
        envvar=env_vars.CONNECTION_ID_ENV_VAR,
        help=f"""Connection ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.CONNECTION_ID_ENV_VAR}.""",
    )


EXPERIMENT_REPOSITORY_ID = click.option(
    "--experiment-repository-id",
    required=True,
    type=uuid.UUID,
    envvar=env_vars.EXPERIMENT_REPOSITORY_ID_ENV_VAR,
    help=f"""Experiment Repository ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.EXPERIMENT_REPOSITORY_ID_ENV_VAR}.""",
)

DATA_REPOSITORY_ID = click.option(
    "--data-repository-id",
    type=uuid.UUID,
    required=True,
    envvar=env_vars.DATA_REPOSITORY_ID_ENV_VAR,
    help=f"""Data Repository ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.DATA_REPOSITORY_ID_ENV_VAR}.""",
)

DATA_REPOSITORY_ID_OPTIONAL = click.option(
    "--data-repository-id",
    type=uuid.UUID,
    required=False,
    envvar=env_vars.DATA_REPOSITORY_ID_ENV_VAR,
    help=f"""Data Repository ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.DATA_REPOSITORY_ID_ENV_VAR}.""",
)

CODE_REPOSITORY_ID = click.option(
    "--code-repository-id",
    required=True,
    type=uuid.UUID,
    envvar=env_vars.CODE_REPOSITORY_ID_ENV_VAR,
    help=f"""Code Repository ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.CODE_REPOSITORY_ID_ENV_VAR}.""",
)


def get_run_id(required: bool):
    return click.option(
        "--run-id",
        required=required,
        default=None,
        type=uuid.UUID,
        envvar=env_vars.RUN_ID_ENV_VAR,
        help=f"""Run ID on Mantik.

    If not specified, it is inferred from the environment variable
{env_vars.RUN_ID_ENV_VAR}.""",
    )


def get_target_dir_option(help_option: str):
    return click.option(
        "--target-dir",
        type=click.Path(path_type=str, exists=True),
        default="./",
        help=help_option,
        show_default=True,
    )


DATA_TARGET_DIR = click.option(
    "--data-target-dir",
    type=click.Path(path_type=str),
    default="data",
    help="Relative path to directory where the data will be stored (from code root)",  # noqa E501
    show_default=True,
    envvar=env_vars.TARGET_DIR_ENV_VAR,
)


def get_target_dir_option_required(help_option: str):
    return click.option(
        "--target-dir",
        type=click.Path(path_type=str),
        help=help_option,
        required=True,
        envvar=env_vars.TARGET_DIR_ENV_VAR,
    )


def get_name_option(required: bool, help_option: str):
    return click.option(
        "--name",
        required=required,
        default=None,
        type=str,
        help=help_option,
    )



import io
import logging
import pathlib
import typing as t
import uuid

import fs.base
import fs.copy
import fs.errors
import fs.ftpfs
import pyunicore.credentials
import pyunicore.uftpfs as uc_uftpfs

import mantik.authentication.auth
import mantik.config.core as core
import mantik.remote_file_service.abstract_file_service as abstract_file_service  # noqa
import mantik.utils.credentials as _credentials

logger = logging.getLogger()

FilePath = abstract_file_service.FilePath


def join(path_1: FilePath, path_2: FilePath) -> FilePath:
    return str(pathlib.Path(path_1).joinpath(pathlib.Path(path_2)))


class UnicoreFileService(abstract_file_service.AbstractFileService):
    """Client that allows transferring files between unicore and
    the local machine.

    To specify that a path is remote, start the path with `remote:`
    E.g. `client.copy_file("/path/local.file", "remote:/path/local.file")`
    This would upload the file from local to unicore.

    Permission handling is not supported here, as it's not implemented by
    pyunicore. This includes viewing the file mode, and file owner.
    """

    def _get_remote_fs(self) -> fs.base.FS:
        return self.remote_fs

    def __init__(
        self,
        username: str,
        password: str,
        auth_url: str,
        remote_base_path: str = "/",
        local_base_path: str = ".",
    ):
        self.username = username
        self.remote_fs = uc_uftpfs.UFTPFS(
            auth_url=auth_url,
            creds=pyunicore.credentials.UsernamePassword(username, password),
            base_path=remote_base_path,
        )
        super().__init__(local_base_path=local_base_path)

    def copy_file(
        self,
        source: FilePath,
        target: FilePath,
    ) -> abstract_file_service.File:
        (
            source_fs,
            relative_source_path,
        ) = self._get_current_fs_and_relative_filepath(source)
        (
            target_fs,
            relative_target_path,
        ) = self._get_current_fs_and_relative_filepath(target)

        fs.copy.copy_file(
            src_fs=source_fs,
            src_path=relative_source_path,
            dst_fs=target_fs,
            dst_path=relative_target_path,
        )

        return self._get_file_details(target)

    def remove_file(self, target=FilePath) -> None:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.remove(relative_path)

    def create_file_if_not_exists(
        self, target=FilePath
    ) -> abstract_file_service.File:
        if not self.exists(target):
            if self.is_remote(target):
                with io.BytesIO() as file:
                    self.remote_fs.upload(self.localise_path(target), file)
            else:
                self._get_fs(target).create(target)
        return self._get_file_details(target)

    def _get_file_details(self, path: FilePath) -> abstract_file_service.File:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            path
        )

        return abstract_file_service.File(
            path=path,
            is_remote=self.is_remote(path),
            metadata=abstract_file_service.FileMeta(
                owner=self.user,
                last_changed=current_fs.getmodified(relative_path),
                mode="Unknown",  # Note: Not supported by pyunicore
            ),
        )

    def _get_directory_details(
        self, path: FilePath, fill_children: bool = False
    ) -> abstract_file_service.Directory:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            path
        )

        if fill_children:
            children = []
            directory_contents = current_fs.listdir(relative_path)

            for item in directory_contents:
                if current_fs.getdetails(join(relative_path, item)).is_dir:
                    children.append(
                        self._get_directory_details(join(path, item))
                    )
                else:
                    children.append(self._get_file_details(join(path, item)))

        else:
            children = []

        return abstract_file_service.Directory(
            path=path,
            is_remote=self.is_remote(path),
            metadata=abstract_file_service.FileMeta(
                owner="Unknown",  # Note: Not supported by pyunicore
                last_changed=current_fs.getmodified(relative_path),
                mode="Unknown",  # Note: Not supported by pyunicore
            ),
            children=children,
        )

    def list_directory(
        self, target: FilePath
    ) -> t.List[
        t.Union[abstract_file_service.Directory, abstract_file_service.File]
    ]:
        return self._get_directory_details(
            path=target, fill_children=True
        ).children

    def create_directory(
        self, target: FilePath
    ) -> abstract_file_service.Directory:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.makedir(relative_path)

        return self._get_directory_details(path=target, fill_children=False)

    def remove_directory(self, target: FilePath) -> None:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.removetree(relative_path)

    def copy_directory(
        self, source: FilePath, target: FilePath
    ) -> abstract_file_service.Directory:
        self._copy_dir(source, target)
        return self._get_directory_details(path=target, fill_children=True)

    @property
    def user(self) -> str:
        return self.username

    def change_permissions(
        self, target: FilePath, new_permissions: abstract_file_service.FileMeta
    ) -> None:
        # Note: Not supported by pyunicore
        raise NotImplementedError()

    def exists(self, target=FilePath) -> bool:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        try:
            return current_fs.exists(relative_path)
        except fs.errors.PermissionDenied:
            return False

    def _copy_dir(
        self,
        source_path: FilePath,
        target_path: FilePath,
    ):
        """
        Function which copies one directory to another, compatible with
        pyunicore uftpfs.

        This custom function is needed because of an incompatability
        between pyfilesystem and pyunicore uftpfs. Upon using the default
        copy dir, the filesystem is checked for the existence of the target dir.
        If the target dir doesn't exist, and the target filesystem is uftpfs,
        this results in a permission denied rather than 'does not exist',
        which crashes the request.
        """
        if not self.exists(target_path):
            logger.info(f"mkdir: {target_path}")
            self.create_directory(target_path)

        (
            source_fs,
            relative_source_path,
        ) = self._get_current_fs_and_relative_filepath(source_path)

        for current_item in source_fs.scandir(relative_source_path):
            current_source_file_path = join(source_path, current_item.name)
            current_target_file_path = join(target_path, current_item.name)
            if current_item.is_dir:
                self._copy_dir(
                    current_source_file_path, current_target_file_path
                )
            else:
                logger.info(
                    f"{current_source_file_path} -> {current_target_file_path}"
                )
                self.copy_file(
                    current_source_file_path, current_target_file_path
                )

    @classmethod
    def from_env(
        cls, connection_id: t.Optional[uuid.UUID] = None
    ) -> "UnicoreFileService":
        credentials = _credentials.HpcRestApi.from_unicore_env_vars(
            connection_id
        )
        auth_url = mantik.utils.env.get_required_env_var(
            core._UNICORE_AUTH_SERVER_URL_ENV_VAR
        )

        remote_base_path = (
            mantik.utils.env.get_optional_env_var(
                abstract_file_service.REMOTE_FS_BASE_PATH_ENV_VAR
            )
            or "/"
        )
        local_base_path = (
            mantik.utils.env.get_optional_env_var(
                abstract_file_service.LOCAL_FS_BASE_PATH_ENV_VAR
            )
            or "."
        )

        return cls(
            username=credentials.username,
            password=credentials.password,
            auth_url=auth_url,
            remote_base_path=remote_base_path,
            local_base_path=local_base_path,
        )

    @classmethod
    def is_remote(cls, path: FilePath) -> bool:
        return path.startswith("remote:")

    @classmethod
    def localise_path(cls, path: FilePath) -> FilePath:
        return path.replace("remote:", "")



import json
import logging
import typing as t

import requests

import mantik.utils as utils
import mantik.utils.mantik_api.credentials as credentials

_MANTIK_API_URL_ENV_VAR = "MANTIK_API_URL"
_MANTIK_API_SUBDOMAIN = "api"
_DEFAULT_MANTIK_API_URL = f"https://{_MANTIK_API_SUBDOMAIN}.cloud.mantik.ai"

_TOKENS_API_PATH_PREFIX = "/mantik/tokens"
MANTIK_API_CREATE_TOKEN_API_PATH = f"{_TOKENS_API_PATH_PREFIX}/create"
MANTIK_API_REFRESH_TOKEN_API_PATH = f"{_TOKENS_API_PATH_PREFIX}/refresh"

logger = logging.getLogger(__name__)


def create_tokens(
    access_credentials: t.Optional[credentials.Credentials] = None,
) -> t.Dict:
    """Get tokens from the Mantik API."""
    if access_credentials is None:
        access_credentials = credentials.Credentials.from_env()
    response = send_request_to_mantik_api(
        method="POST",
        data=access_credentials.to_dict(),
        url_endpoint=MANTIK_API_CREATE_TOKEN_API_PATH,
    )
    return response.json()


def refresh_tokens(
    refresh_token: str,
    access_credentials: t.Optional[credentials.Credentials] = None,
) -> t.Dict:
    """Get tokens from the Mantik API."""
    if access_credentials is None:
        access_credentials = credentials.Credentials.from_env()
    data = {
        **access_credentials.to_dict(include_password=False),
        "refresh_token": refresh_token,
    }
    response = send_request_to_mantik_api(
        method="POST", data=data, url_endpoint=MANTIK_API_REFRESH_TOKEN_API_PATH
    )
    return response.json()


def send_request_to_mantik_api(
    method: str,
    data: t.Union[dict, str],
    url_endpoint: str,
    token: t.Optional[str] = None,
    query_params: t.Optional[dict] = None,
) -> requests.Response:
    url = _create_url(url_endpoint)
    header = (
        {
            "Authorization": f"Bearer {token}",
            "Accept": "application/json",
        }
        if token
        else None
    )
    request = {
        "url": url,
        "json": data,
        "headers": header,
        "params": query_params,
    }
    try:
        response = requests.request(method, **request)
        response.raise_for_status()
    except requests.HTTPError as e:
        logger.exception(
            "Call to Mantik API %s with data %s failed",
            url,
            data,
            exc_info=True,
        )
        try:
            message = e.response.json()["detail"]
        except (KeyError, json.decoder.JSONDecodeError):
            message = e.response.text
        raise requests.exceptions.HTTPError(
            f"Request to Mantik API failed: {message}",
            response=e.response,
        ) from e
    else:
        return response


def _create_url(endpoint: str) -> str:
    url = _get_base_url()
    url = utils.urls.remove_double_slashes_from_path(
        f"{url}{endpoint}", ensure_https=False
    )
    return url


def _get_base_url() -> str:
    """Gets the base URL for the mantik API.

    Notes
    -----
    Tries to get the base URL from (in order)
    `MANTIK_API_URL` env var

    Defaults to `https://api.cloud.mantik.ai`


    """
    mantik_api_url = utils.env.get_optional_env_var(_MANTIK_API_URL_ENV_VAR)
    if mantik_api_url is not None:
        logger.debug("Base URL for mantik API is set to %s", mantik_api_url)
        return mantik_api_url

    logger.debug(
        "Base URL for mantik API is set to %s", _DEFAULT_MANTIK_API_URL
    )
    return _DEFAULT_MANTIK_API_URL



import logging
import pathlib
import typing as t
import uuid

import click

import mantik.cli._options as main_options
import mantik.cli.runs._options as _options
import mantik.cli.runs.runs as runs
import mantik.cli.utils as utils
import mantik.runs.local as local_runs
import mantik.utils

logger = logging.getLogger(__name__)


@runs.cli.command("local")
@_options.MLPROJECT_PATH
@main_options.get_name_option(required=True, help_option="Name of the Run.")
@_options.ENTRY_POINT
@main_options.PROJECT_ID
@main_options.DATA_REPOSITORY_ID_OPTIONAL
@main_options.DATA_TARGET_DIR
@click.option(
    "--data-commit",
    required=False,
    type=str,
    help="Commit hash of data repo to checkout. Takes precedence over data-branch",  # noqa E501
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_COMMIT_ENV_VAR,
    default=None,
)
@click.option(
    "--data-branch",
    required=False,
    type=str,
    help="Branch of data repo to checkout. Defaults main branch.",
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_BRANCH_ENV_VAR,
    default=None,
)
@main_options.EXPERIMENT_REPOSITORY_ID
@main_options.CODE_REPOSITORY_ID
@_options.BRANCH
@_options.COMMIT
@click.option(
    "--env-manager",
    type=str,
    default="local",
    help="Determines what environment manager is used for installing"
    " dependencies during runs (local, virtualenv or conda).",
)
@_options.PARAMETER
@main_options.VERBOSE
def run_project(
    name: str,
    mlproject_path: pathlib.Path,
    entry_point: str,
    parameter: t.List[str],
    verbose: bool,  # noqa
    project_id: uuid.UUID,
    data_repository_id: t.Optional[uuid.UUID],
    data_target_dir: t.Optional[str],
    data_branch: t.Optional[str],
    data_commit: t.Optional[str],
    experiment_repository_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    branch: t.Optional[str],
    commit: t.Optional[str],
    env_manager: t.Optional[str],
) -> None:
    """Run an MLflow project locally and save the results in Mantik API

    Note that `MLPROJECT_PATH` is the relative path to the MLflow project file
    with your Code Repository as root.

    Remember that when you execute a run, the code is retrieved from your
    remote Git Code Repository. So make sure to commit and push your
    changes before executing a run!

    To find the respective required IDs make sure to check Mantik's UI

    """

    _options.check_commit_or_branch(branch=branch, commit=commit, logger=logger)

    logger.debug("Parsing MLflow entry point parameters")
    parameters = utils.dict_from_list(parameter)

    local_runs.start_local_run(
        project_id=project_id,
        name=name,
        experiment_repository_id=experiment_repository_id,
        data_repository_id=data_repository_id,
        code_repository_id=code_repository_id,
        branch=branch,
        commit=commit,
        data_branch=data_branch,
        data_commit=data_commit,
        mlflow_mlproject_file_path=mlproject_path.as_posix(),
        entry_point=entry_point,
        mlflow_parameters=parameters,
        data_target_dir=data_target_dir,
        env_manager=env_manager,
    )
    click.echo("Done!")



import pathlib
import typing as t

import mantik.config as config
import mantik.utils.credentials as _credentials

MLFLOW_ENV_TEST_ENV_VAR = "MLFLOW_TESTING_MANTIK_TEST_VAR"

_BASE_ENV_VARS = [
    config.core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
    MLFLOW_ENV_TEST_ENV_VAR,
]
ALL_UNICORE_ENV_VARS = [
    _credentials.UNICORE_USERNAME_ENV_VAR,
    _credentials.UNICORE_PASSWORD_ENV_VAR,
    config.core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
    MLFLOW_ENV_TEST_ENV_VAR,
]
ALL_FIRECREST_ENV_VARS = [
    _credentials.FIRECREST_CLIENT_ID_ENV_VAR,
    _credentials.FIRECREST_CLIENT_SECRET_ENV_VAR,
    config.core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR,
    MLFLOW_ENV_TEST_ENV_VAR,
]

DEFAULT_ENV_VAR_VALUE = "test-val"

EXISTING_FILE = (
    pathlib.Path(__file__).parent / "../../../resources/test-project/recipe.def"
).as_posix()


def _create_apptainer_environment(
    path: pathlib.Path,
    type: str = None,
    variables: t.Optional[t.Dict] = None,
    modules: t.Optional[t.List] = None,
    pre_run_command_on_login_node: t.Optional[list] = None,
    pre_run_command_on_compute_node: t.Optional[list] = None,
    post_run_command_on_compute_node: t.Optional[list] = None,
    post_run_command_on_login_node: t.Optional[list] = None,
    include_mlflow_env_vars: bool = True,
) -> config.environment.Environment:
    if type is None:
        type = "local"
    if include_mlflow_env_vars:
        variables = _include_mlflow_env_vars(variables)

    return config.environment.Environment(
        execution=config.executable.Apptainer(
            path=path,
            type=type,
        ),
        variables=variables,
        modules=modules,
        pre_run_command_on_login_node=pre_run_command_on_login_node,
        pre_run_command_on_compute_node=pre_run_command_on_compute_node,
        post_run_command_on_compute_node=post_run_command_on_compute_node,
        post_run_command_on_login_node=post_run_command_on_login_node,
    )


def _create_python_environment(
    path: pathlib.Path,
    variables: t.Optional[t.Dict] = None,
    modules: t.Optional[t.List] = None,
    pre_run_command_on_login_node: t.Optional[list] = None,
    pre_run_command_on_compute_node: t.Optional[list] = None,
    post_run_command_on_compute_node: t.Optional[list] = None,
    post_run_command_on_login_node: t.Optional[list] = None,
    include_mlflow_env_vars: bool = True,
) -> config.environment.Environment:
    if include_mlflow_env_vars:
        variables = _include_mlflow_env_vars(variables)

    return config.environment.Environment(
        config.executable.Python(
            path=path,
        ),
        variables=variables,
        modules=modules,
        pre_run_command_on_login_node=pre_run_command_on_login_node,
        pre_run_command_on_compute_node=pre_run_command_on_compute_node,
        post_run_command_on_compute_node=post_run_command_on_compute_node,
        post_run_command_on_login_node=post_run_command_on_login_node,
    )


def _include_mlflow_env_vars(env_variables: t.Optional[t.Dict]) -> t.Dict:
    mlflow_env_vars = {
        MLFLOW_ENV_TEST_ENV_VAR: DEFAULT_ENV_VAR_VALUE,
    }
    if env_variables is None:
        env_variables = mlflow_env_vars
    else:
        env_variables = {**env_variables, **mlflow_env_vars}
    return env_variables


def _create_config_with_unicore(
    resources: config.resources.Resources,
    unicore_api_url: str = "test-url",
    user: str = "test-val",
    password: str = "test-val",
    project: str = "test-val",
    env: t.Optional[config.environment.Environment] = None,
) -> config.core.Config:
    env = _get_env_vars(env)
    return config.core.Config(
        unicore_api_url=unicore_api_url,
        user=user,
        password=password,
        project=project,
        resources=resources,
        environment=env,
    )


def _get_env_vars(
    env: t.Optional[config.environment.Environment] = None,
) -> config.environment.Environment:
    if env is None:
        env = config.environment.Environment()
    env.variables = _include_mlflow_env_vars(env.variables)
    return env


def _create_config_with_firecrest(
    resources: config.resources.Resources,
    api_url: str = "test-api-url",
    token_url: str = "test-token-url",
    machine: str = "test-machine",
    user: str = "test-val",
    password: str = "test-val",
    project: str = "test-val",
    env: t.Optional[config.environment.Environment] = None,
) -> config.core.Config:
    env = _get_env_vars(env)
    return config.core.Config(
        firecrest=config.firecrest.Firecrest(
            api_url=api_url, token_url=token_url, machine=machine
        ),
        user=user,
        password=password,
        project=project,
        resources=resources,
        environment=env,
    )



import dataclasses
import logging
import typing as t
import uuid

import fs.copy
import fs_s3fs
import jose.jwt

import mantik.authentication
import mantik.remote_file_service.abstract_file_service as afs  # noqa
import mantik.utils.mantik_api.connection

logger = logging.getLogger()

FilePath = afs.FilePath


class S3FileService(afs.AbstractFileService):
    """Client that allows transferring files between s3 and
    the local machine.

    To specify that a path is remote, start the path with `s3://`
    E.g. `client.copy_file("/path/local.file", "s3://path/local.file")`
    This would upload the file from local to s3.
    """

    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        local_base_path: str = ".",
    ):
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        super().__init__(local_base_path=local_base_path)

    def copy_file(
        self,
        source: FilePath,
        target: FilePath,
    ) -> afs.File:
        (
            source_fs,
            relative_source_path,
        ) = self._get_current_fs_and_relative_filepath(source)
        (
            target_fs,
            relative_target_path,
        ) = self._get_current_fs_and_relative_filepath(target)

        fs.copy.copy_file(
            src_fs=source_fs,
            src_path=relative_source_path,
            dst_fs=target_fs,
            dst_path=relative_target_path,
        )

        return self._get_file_details(target)

    def remove_file(self, target=FilePath) -> None:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.remove(relative_path)

    def create_file_if_not_exists(self, target=FilePath) -> afs.File:
        current_fs, local_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.create(local_path, wipe=False)
        return self._get_file_details(target)

    def list_directory(
        self, target: FilePath
    ) -> t.List[t.Union[afs.Directory, afs.File]]:
        return self._get_directory_details(
            path=target, fill_children=True
        ).children

    def create_directory(self, target: FilePath) -> afs.Directory:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.makedir(relative_path)

        return self._get_directory_details(path=target, fill_children=False)

    def remove_directory(self, target: FilePath) -> None:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        current_fs.removetree(relative_path)

    def copy_directory(
        self, source: FilePath, target: FilePath
    ) -> afs.Directory:
        src_fs, src_path = self._get_current_fs_and_relative_filepath(source)
        dst_fs, dst_path = self._get_current_fs_and_relative_filepath(target)

        fs.copy.copy_dir(
            src_fs=src_fs, src_path=src_path, dst_fs=dst_fs, dst_path=dst_path
        )
        return self._get_directory_details(path=target, fill_children=True)

    @property
    def user(self) -> str:
        return self.aws_access_key_id

    def change_permissions(
        self, target: FilePath, new_permissions: afs.FileMeta
    ) -> None:
        raise NotImplementedError()

    def exists(self, target=FilePath) -> bool:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            target
        )
        try:
            return current_fs.exists(relative_path)
        except fs.errors.PermissionDenied:
            return False

    def _get_directory_details(
        self, path: FilePath, fill_children: bool = False
    ) -> afs.Directory:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            path
        )

        if fill_children:
            children = []
            directory_contents = current_fs.listdir(relative_path)
            for item in directory_contents:
                if current_fs.getdetails(
                    fs.path.combine(relative_path, item)
                ).is_dir:
                    children.append(
                        self._get_directory_details(fs.path.combine(path, item))
                    )
                else:
                    children.append(
                        self._get_file_details(fs.path.combine(path, item))
                    )

        else:
            children = []

        return afs.Directory(
            path=path,
            is_remote=self.is_remote(path),
            metadata=afs.FileMeta(
                owner="Unknown",
                last_changed=current_fs.getmodified(relative_path),
                mode="Unknown",
            ),
            children=children,
        )

    def _get_file_details(self, path: FilePath) -> afs.File:
        current_fs, relative_path = self._get_current_fs_and_relative_filepath(
            path
        )

        return afs.File(
            path=path,
            is_remote=self.is_remote(path),
            metadata=afs.FileMeta(
                owner=self.user,
                last_changed=current_fs.getmodified(relative_path),
                mode="Unknown",  # Note: Not supported by python fs
            ),
        )

    def _get_remote_fs(self, bucket_name: str) -> fs_s3fs.S3FS:
        return fs_s3fs.S3FS(
            bucket_name=bucket_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )

    @classmethod
    def from_env(
        cls, connection_id: t.Optional[uuid.UUID] = None
    ) -> "S3FileService":
        s3_credentials = S3Credentials.get_credentials(
            connection_id=connection_id
        )

        local_base_path = (
            mantik.utils.env.get_optional_env_var(
                afs.LOCAL_FS_BASE_PATH_ENV_VAR
            )
            or "."
        )

        return cls(
            aws_access_key_id=s3_credentials.aws_access_key_id,
            aws_secret_access_key=s3_credentials.aws_secret_access_key,
            local_base_path=local_base_path,
        )

    @classmethod
    def is_remote(cls, path: FilePath) -> bool:
        """Checks whether a given url is an S3 url."""
        return path.startswith("s3://")

    @classmethod
    def localise_path(cls, path: FilePath) -> FilePath:
        """Strips s3 prefix and bucket name from S3 url."""
        bucket_name = cls.get_bucket_name(path)
        remainder = path.replace(f"s3://{bucket_name}", "")
        if remainder.startswith("/"):
            return remainder[1::]
        return remainder

    @classmethod
    def get_bucket_name(cls, path: FilePath) -> str:
        """Extracts bucket name from an S3 url
        `s3://<bucket-name>/resource/...`"""
        if not cls.is_remote(path):
            raise ValueError("Not an s3 url.")

        path_without_prefix = path.replace("s3://", "")
        return path_without_prefix.split("/")[0]

    def _get_fs_nargs(self, path: FilePath) -> dict:
        return {"bucket_name": self.get_bucket_name(path)}


_USERNAME_ENV_VAR = "AWS_ACCESS_KEY_ID"
_PASSWORD_ENV_VAR = "AWS_SECRET_ACCESS_KEY"


@dataclasses.dataclass
class S3Credentials:
    aws_access_key_id: str
    aws_secret_access_key: str

    @classmethod
    def get_credentials(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
    ) -> "S3Credentials":
        if connection_id:
            return cls._credentials_from_api(connection_id=connection_id)
        return cls._credentials_from_env_vars()

    @classmethod
    def _credentials_from_api(
        cls,
        connection_id: uuid.UUID,
    ) -> "S3Credentials":
        access_token = mantik.authentication.auth.get_valid_access_token()
        user_id = _get_sub_from_token(access_token)
        connection = mantik.utils.mantik_api.connection.get(
            user_id=uuid.UUID(user_id),
            connection_id=connection_id,
            token=access_token,
        )
        return cls(
            aws_access_key_id=connection.login_name,
            aws_secret_access_key=connection.password,
        )

    @classmethod
    def _credentials_from_env_vars(cls) -> "S3Credentials":
        username = mantik.utils.env.get_required_env_var(_USERNAME_ENV_VAR)
        password = mantik.utils.env.get_required_env_var(_PASSWORD_ENV_VAR)
        return cls(aws_access_key_id=username, aws_secret_access_key=password)


def _get_sub_from_token(token: str):
    # The `sub` field of the token claims contains the user UUID.
    return jose.jwt.get_unverified_claims(token)["sub"]



import dataclasses
import json
import unittest.mock
import uuid

import pytest
import starlette.status

import mantik_compute_backend.api as api
import mantik_compute_backend.models as models

TEST_UUID = uuid.uuid4()


@dataclasses.dataclass
class FakeSubmittedRun:
    # MLflow returns run IDs as UUID string (`uuid.hex`)
    run_id: str = TEST_UUID.hex
    job_id: str = "1"


@pytest.mark.parametrize(
    ("headers", "expected"),
    [
        ({}, 401),
        (
            {"Authorization": "test-invalid-token"},
            401,
        ),
        (
            {"Authorization": "Bearer test-invalid-token"},
            401,
        ),
        (
            {"Authorization": "Bearer test-valid-token"},
            201,
        ),
    ],
)
@unittest.mock.patch(
    "mlflow.projects.run",
    return_value=FakeSubmittedRun(),
)
def test_submit_run(
    mock_mlflow_run, client, zipped_unicore_content, headers, expected
):
    response = client.post(
        f"{api.SUBMIT_PATH}/0",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "bar",
            "hpc_api_password": "baz",
            "compute_budget_account": "empty",
            "mlflow_tracking_uri": "foo.bar",
            "mlflow_tracking_token": "abcdefghijk",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers=headers,
    )

    assert response.status_code == expected, response.text

    if expected == 201:
        expected_json = {
            "experiment_id": 0,
            "run_id": str(TEST_UUID),
            "unicore_job_id": "1",
        }
        result = response.json()

        assert result == expected_json

        mock_mlflow_run.assert_called()


@unittest.mock.patch(
    "mlflow.projects.run",
    return_value=FakeSubmittedRun(),
)
def test_submit_run_invalid_backend_config_type(
    mock_mlflow_run, client, zipped_unicore_content
):
    unsupported_format = ".yamml"
    response = client.post(
        f"{api.SUBMIT_PATH}/0",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "bar",
            "hpc_api_password": "baz",
            "compute_budget_account": "empty",
            "mlflow_tracking_uri": "foo.bar",
            "mlflow_tracking_token": "abcdefghijk",
            "compute_backend_config": f"compute-backend-config{unsupported_format}",  # noqa: E501
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )

    error_message = response.json()["message"]
    mock_mlflow_run.assert_not_called()
    assert response.status_code == 400
    assert (
        error_message == "While submitting the job, "
        "this configuration error occurred: "
        f"The given file type '{unsupported_format}'"
        " is not supported for the config,"
        " the supported ones are: '.json', '.yml', '.yaml'."
    )


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    return_value=models.SubmitRunResponse(
        experiment_id=0, run_id=uuid.uuid4(), unicore_job_id="0"
    ),
)
@unittest.mock.patch("tempfile.TemporaryFile")
def test_memory_freed_when_success(
    mock_temporary_file,
    mock_handle_submit_run_request,
    client,
    zipped_unicore_content,  # noqa
):
    response = client.post(
        f"{api.SUBMIT_PATH}/0",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "bar",
            "hpc_api_password": "baz",
            "compute_budget_account": "empty",
            "mlflow_tracking_uri": "foo.bar",
            "mlflow_tracking_token": "abcdefghijk",
            "compute_backend_config": "backend-config.yaml",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 201
    mock_temporary_file.assert_called()
    mock_temporary_file().write.assert_called()
    mock_temporary_file().seek.assert_called()
    mock_temporary_file().close.assert_called()
    mock_handle_submit_run_request.assert_called()


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{"return_value.raiseError.side_effect": Exception},
)
@unittest.mock.patch("tempfile.TemporaryFile")
def test_memory_freed_when_exception(
    mock_temp_file, mocked_run, client_suppressing_raise, zipped_unicore_content
):
    mocked_run.side_effect = Exception
    response = client_suppressing_raise.post(
        f"{api.SUBMIT_PATH}/0",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "bar",
            "hpc_api_password": "baz",
            "compute_budget_account": "empty",
            "mlflow_tracking_uri": "foo.bar",
            "mlflow_tracking_token": "abcdefghijk",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )

    assert response.status_code == 500
    mock_temp_file().close.assert_called()


def test_submit_too_large_file(
    client_with_small_size_limitation, zipped_unicore_content
):
    response = client_with_small_size_limitation.post(
        f"{api.SUBMIT_PATH}/0",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "bar",
            "hpc_api_password": "baz",
            "compute_budget_account": "empty",
            "mlflow_tracking_uri": "foo.bar",
            "mlflow_tracking_token": "abcdefghijk",
        },
        files={"mlproject_zip": zipped_unicore_content},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert (
        response.status_code
        == starlette.status.HTTP_413_REQUEST_ENTITY_TOO_LARGE
    )



import contextlib
import os
import pathlib
import typing as t

import pytest

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture
def env_vars_set():
    @contextlib.contextmanager
    def wrapped(env_vars: t.Dict[str, t.Any]):
        for key, value in env_vars.items():
            os.environ[key] = value
        yield
        for key in env_vars:
            try:
                os.environ.pop(key)
            except KeyError:
                pass

    return wrapped


@pytest.fixture
def expect_raise_if_exception() -> (
    t.Callable[[t.Any], contextlib.AbstractContextManager]
):
    def expect_can_be_error(
        expected: t.Any,
    ) -> contextlib.AbstractContextManager:
        return (
            pytest.raises(type(expected))
            if isinstance(expected, Exception)
            else contextlib.nullcontext()
        )

    return expect_can_be_error


@pytest.fixture()
def example_project_path() -> pathlib.Path:
    return FILE_DIR / "resources/test-project"



import logging
import pathlib
import typing as t

import mantik.config.core as core
import mantik.config.exceptions as exceptions
import mantik.config.executable as executable
import mantik.utils.unicore.upload as upload

logger = logging.getLogger(__name__)


class ProjectValidator:
    def __init__(
        self,
        mlproject_path: t.Union[pathlib.Path, str],
        config: t.Union[pathlib.Path, str, dict],
        mlflow_parameters: t.Dict,
        entry_point: str,
        logger_level: int = logging.WARNING,
    ):
        self.mlproject_path = pathlib.Path(mlproject_path)
        if isinstance(config, dict):
            self._config = core.Config._from_dict(config)
        else:
            self.config_path = pathlib.Path(config)
        self.mlflow_parameters = mlflow_parameters
        self.entry_point = entry_point
        logger.setLevel(logger_level)

        self._config = None

    @property
    def config(self) -> core.Config:
        if self._config is None:
            self._config = core.Config.from_filepath(self.config_absolute_path)
        return self._config

    @property
    def config_absolute_path(self) -> pathlib.Path:
        return (
            self.config_path
            if self.config_path.is_absolute()
            else self.mlproject_path / self.config_path
        )

    @property
    def config_relative_path(self) -> pathlib.Path:
        return (
            self.config_path
            if not self.config_path.is_absolute()
            else self.config_path.relative_to(self.mlproject_path)
        )

    def validate(self) -> None:
        logger.debug(
            "Validating MLproject at %s and backend config at %s",
            self.mlproject_path.as_posix(),
            self.config_absolute_path.as_posix(),
        )
        if not self.mlproject_path.is_absolute():
            raise exceptions.ConfigValidationError(
                "ML project path must be an absolute path, "
                f"but {self.mlproject_path.as_posix()!r} was given."
            )
        if not self.mlproject_path.is_dir():
            raise exceptions.ConfigValidationError(
                "ML project path not found at "
                f"{self.mlproject_path.as_posix()!r}, "
                "check that the given path is correct."
            )
        if not self.config_absolute_path.is_file():
            raise exceptions.ConfigValidationError(
                "Config not found at "
                f"{self.config_absolute_path.as_posix()!r}, "
                "check that the given path is correct."
            )
        if not (self.mlproject_path in self.config_absolute_path.parents):
            raise exceptions.ConfigValidationError(
                "Config file is not in the ML project directory, "
                "check that the given paths are correct:\n"
                f"Config file: {self.config_absolute_path.as_posix()!r}\n"
                f"ML project directory: {self.mlproject_path.as_posix()!r}"
            )
        if self.config.execution_environment_given():
            self._validate_execution_path(self.config.environment.execution)
        logger.debug("Validating MLproject configuration.")
        self._validate_ml_project_file()
        self._validate_config_not_in_exclude()

    def _validate_execution_path(self, execution: executable.Execution):
        if execution.path_has_to_be_checked:
            execution_absolute_path = self.mlproject_path / execution.path
            if not execution_absolute_path.is_file():
                raise exceptions.ConfigValidationError(
                    f"The path {execution_absolute_path.as_posix()!r} given "
                    "as execution environment was not found, "
                    "check that the given path is correct. "
                    "The path must be relative to the ML project path."
                )

    def _validate_ml_project_file(self) -> None:
        try:
            import mlflow
        except ModuleNotFoundError:
            logger.warning(
                (
                    "Cannot validate MLproject file since mlflow package "
                    "is not installed. Either install the supported version "
                    'via `pip install "mantik[mlflow]"` or install a '
                    "desired MLflow version manually."
                ),
                exc_info=True,
            )
        else:
            mlflow.projects.utils.fetch_and_validate_project(
                uri=self.mlproject_path.as_posix(),
                version=None,
                entry_point=self.entry_point,
                parameters=self.mlflow_parameters,
            )

    def _validate_config_not_in_exclude(self) -> None:
        if upload.file_matches_exclude_pattern(
            path=self.config_absolute_path,
            project_dir=self.mlproject_path,
            exclude=self.config.files_to_exclude,
        ):
            raise exceptions.ConfigValidationError(
                f"Config file '{self.config_path.name}' cannot be excluded, "
                f"check config 'Exclude' section"
            )



import os
import pathlib

import pytest

import mantik.config as config
import mantik.utils as utils
import mantik.utils.credentials as _credentials

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture
def required_config_env_vars():
    return {
        _credentials.UNICORE_USERNAME_ENV_VAR: "test-user",
        _credentials.UNICORE_PASSWORD_ENV_VAR: "test-password",
        config.core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR: "test-project",
    }


@pytest.fixture(scope="session")
def mlproject_path() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent
        / "../../../../tests/resources/test-project"
    )


@pytest.fixture(scope="session")
def invalid_config_type() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent
        / "../../../../tests/resources/broken-project/compute-backend-config.md"  # noqa: E501
    )


@pytest.fixture(scope="session")
def config_with_errors() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent / "../../../../tests/resources/"
        "test-project/config-with-errors.yaml"
    )


@pytest.fixture(scope="session")
def compute_backend_config_yaml(mlproject_path) -> pathlib.Path:
    """Return the UNICORE config in YAML format.

    Doesn't contain. `Environment` section

    """
    return pathlib.Path(f"{str(mlproject_path)}/compute-backend-config.yaml")


@pytest.fixture(scope="session")
def compute_backend_config_json(mlproject_path) -> pathlib.Path:
    return pathlib.Path(f"{str(mlproject_path)}/compute-backend-config.json")


@pytest.fixture()
def unset_tracking_token_env_var_before_execution():
    if utils.mlflow.TRACKING_TOKEN_ENV_VAR in os.environ:
        del os.environ[utils.mlflow.TRACKING_TOKEN_ENV_VAR]


@pytest.fixture(scope="function")
def example_unicore_config() -> config.core.Config:
    return config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            ),
            variables={"SRUN_CPUS_PER_TASK": 100},
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.py", "*.sif"],
    )


@pytest.fixture()
def example_unicore_config_for_python() -> config.core.Config:
    return config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Python(
                path=pathlib.Path("/venv"),
            )
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )



import logging
import pathlib

import mlflow.projects as projects

FILE_PATH = pathlib.Path(__file__).parent

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


if __name__ == "__main__":
    uri = FILE_PATH
    entry_point = "main"
    version = None
    experiment_name = None
    experiment_id = None
    param_dict = None
    backend = "mantik"
    backend_config = FILE_PATH / "compute-backend-config.json"
    no_conda = True
    storage_dir = None
    run_id = None

    run_submitted = projects.run(
        uri.as_posix(),
        entry_point,
        version,
        experiment_name=experiment_name,
        experiment_id=experiment_id,
        parameters=param_dict,
        backend=backend,
        backend_config=backend_config.as_posix(),
        storage_dir=storage_dir,
        synchronous=True,
        run_id=run_id,
    )

    logger.debug(f"Properties: {run_submitted.properties}")
    content_output = run_submitted.read_file_from_working_directory(
        pathlib.Path("mantik.log")
    )
    logger.debug(f"UNICORE logs: {run_submitted.logs}")
    print(f"Mantik logs: {content_output}")



"""Test that UnicoreErrors are passed to the client."""
import functools
import json
import os
import unittest.mock
import uuid

import firecrest
import requests
import requests.exceptions as request_exceptions

import mantik.testing as testing
import mantik_compute_backend.exceptions as exceptions


def test_config_error_response(client, broken_zipped_content, tmp_path):
    """Test that ConfigErrors are passed to the client."""
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.md",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": broken_zipped_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert response.text == (
        '{"message":"While submitting the job, '
        "this configuration error occurred: "
        "The given file type '.md' "
        "is not supported for the config, the "
        "supported ones are: '.json', '.yml', '.yaml'.\"}"
    )


def test_unicore_http_error_response(
    client, mlproject_path, zipped_unicore_content, tmp_path
):
    """Test that UnicoreErrors are passed to the client."""
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert response.text == (
        '{"message":"Unicore backend error. '
        "Cause: Failed to connect to "
        "https://zam2125.zam.kfa-juelich.de:9112/JUWELS/rest/core "
        '-- check if user and password are correct"}'
    )
    # Note: Experiment ID environment variable is set implicitly here,
    # breaking other tests
    os.environ.pop("MLFLOW_EXPERIMENT_ID")


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{"return_value.raiseError.side_effect": request_exceptions.HTTPError},
)
def test_unicore_error_response(
    mocked_run, client, mlproject_path, zipped_unicore_content, tmp_path
):
    """Test that HTTPErrors are passed to the client."""
    mocked_run.side_effect = request_exceptions.HTTPError("Test")
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert (
        response.text == '{"message":"While submitting the job, '
        'this request error occurred: Test"}'
    )


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{"return_value.raiseError.side_effect": ConnectionError},
)
def test_unicore_connection_error_response(
    mocked_run, client, mlproject_path, zipped_unicore_content, tmp_path
):
    """Test that ConnectionErrors are passed to the client."""
    mocked_run.side_effect = ConnectionError("Test")
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert response.text == (
        '{"message":"While submitting the job, '
        'this connection error occurred: Test"}'
    )


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{"return_value.raiseError.side_effect": firecrest.FirecrestException},
)
def test_firecrest_error_response(
    mocked_run,
    client,
    mlproject_path,
    zipped_firecrest_content,
    tmp_path,
    monkeypatch,
):
    """Test that HTTPErrors are passed to the client."""
    fake_response_object = requests.models.Response()
    fake_response_object.status_code = 400
    mocked_run.side_effect = firecrest.FirecrestException(
        responses=[fake_response_object]
    )
    monkeypatch.setattr(
        firecrest,
        "ClientCredentialsAuth",
        functools.partial(
            testing.firecrest.FakeClientCredentialsAuth,
            login_successful=True,
        ),
    )
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-firecrest-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_firecrest_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert response.text == (
        '{"message":"Firecrest backend error. Cause: last request: 400 None"}'
    )


def test_firecrest_credentials_response(
    client, mlproject_path, zipped_firecrest_content, tmp_path
):
    """Test that HTTPErrors are passed to the client."""
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"print": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-firecrest-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_firecrest_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 401, response.text
    assert response.text == (
        '{"message":"Firecrest backend error. Cause: Failed to connect to '
        "https://auth.cscs.ch/auth/realms/firecrest-clients/protocol/openid-connect/token "  # noqa E501
        '-- check if client ID and secret are correct"}'
    )


def test_mlflow_error_response(
    client, mlproject_path, zipped_unicore_content, tmp_path
):
    """Test that MLflow exceptions are passed to the client."""
    experiment_id = 0
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "non-existent-entry-point",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 400, response.text
    assert response.text == (
        '{"message":"While submitting the job, '
        "this mlflow error occurred: Could not "
        "find non-existent-entry-point among "
        "entry points ['main', 'echo', 'multi-line'] or interpret "
        "non-existent-entry-point as a runnable script. "
        "Supported script file "
        "extensions: ['.py', '.sh']\"}"
    )
    # Note: Experiment ID environment variable and Mlflow tracking uri,
    # is set implicitly here,
    # breaking other tests
    os.environ.pop("MLFLOW_EXPERIMENT_ID")
    os.environ.pop("MLFLOW_TRACKING_URI")


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{"return_value.raiseError.side_effect": Exception},
)
def test_custom_500_internal_server_error(
    mocked_run,
    client_suppressing_raise,
    mlproject_path,
    zipped_unicore_content,
    tmp_path,
):
    """Test custom internal server error."""
    mocked_run.side_effect = Exception
    experiment_id = 0
    response = client_suppressing_raise.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    assert response.status_code == 500, response.text
    assert response.text == '{"message":"Mantik Internal Server Error"}'


@unittest.mock.patch(
    "mantik_compute_backend.handle_submit_run.handle_submit_run_request",
    **{
        "return_value.raiseError.side_effect": exceptions.RequestEntityTooLargeException  # noqa
    },
)
def test_file_too_large_exception(
    mocked_run, client, mlproject_path, zipped_unicore_content, tmp_path
):
    """Test custom internal server error."""
    experiment_id = 0
    mocked_run.side_effect = exceptions.RequestEntityTooLargeException()
    response = client.post(
        f"/submit/{experiment_id}",
        data={
            "active_run_id": str(uuid.uuid4().hex),
            "entry_point": "main",
            "mlflow_parameters": json.dumps({"foo": "bar"}),
            "hpc_api_user": "me",
            "hpc_api_password": "bar",
            "compute_budget_account": "empty",
            "compute_backend_config": "compute-backend-config.yaml",
            "mlflow_tracking_uri": (tmp_path / "foo.bar").as_uri(),
            "mlflow_tracking_token": "abc",
        },
        files={"mlproject_zip": zipped_unicore_content.read()},
        headers={"Authorization": "Bearer test-valid-token"},
    )
    print(response.text)
    assert response.status_code == 413, response.text
    assert (
        response.text
        == '{"message":"The files you sent were too large to be handled by '
        "the API. Consider using scp for direct file transfer to remote "
        'compute resources. The maximum allowed size is 100.0 MB."}'
    )



import datetime
import io
import typing as t

import pyunicore.client

import mantik_compute_backend.submitted_run as submitted_run
import mantik_compute_backend.unicore._properties as _properties
import mantik_compute_backend.unicore.job as _job

UNICORE_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
TZ_OFFSET = datetime.timezone(datetime.timedelta(seconds=7200))


class FakeTransport:
    def __init__(self, credential: str = "test_token", oidc: bool = True):
        self.credential = credential
        self.oidc = oidc


class FakeJob:
    def __init__(
        self,
        transport: FakeTransport,
        job_url: str = "test-job",
        properties: t.Dict = None,
        existing_files: t.Dict[str, str] = None,
        will_be_successful: bool = True,
        submission_day: int = 1,
    ):
        self.transport = transport
        self.url = job_url
        self._properties = {
            "status": "SUCCESSFUL",
            "log": [],
            "owner": "owner",
            "siteName": "siteName",
            "consumedTime": {
                "total": 1,
                "queued": 2,
                "stage-in": 3,
                "preCommand": 4,
                "main": 5,
                "postCommand": 6,
                "stage-out": 7,
            },
            "currentTime": _format_datetime_str_with_time_zone(submission_day),
            "submissionTime": _format_datetime_str_with_time_zone(
                submission_day
            ),
            "terminationTime": _format_datetime_str_with_time_zone(
                submission_day + 1
            ),
            "statusMessage": "statusMessage",
            "tags": [],
            "resourceStatus": "resourceStatus",
            "name": "name",
            "exitCode": "0",
            "queue": "queue",
            "submissionPreferences": {},
            "resourceStatusMessage": "N/A",
            "acl": [],
            "batchSystemID": "N/A",
        }

        # Override default properties if any properties given.
        if properties is not None:
            self._properties = {**self._properties, **properties}

        self._existing_files = existing_files or {}
        self._successful = will_be_successful
        self.working_dir = FakeWorkDir()
        self.started = False

    @property
    def properties(self) -> t.Dict:
        return self._properties

    @property
    def job_id(self) -> str:
        return self.url

    def poll(self) -> None:
        self._properties["status"] = (
            "SUCCESSFUL" if self._successful else "FAILED"
        )

    def abort(self) -> None:
        pass

    def start(self) -> None:
        self.started = True


def _format_datetime_str_with_time_zone(day: int) -> str:
    return datetime.datetime(
        2000,
        1,
        day,
        tzinfo=TZ_OFFSET,
    ).strftime(UNICORE_TIMESTAMP_FORMAT)


class FakeWorkDir:
    DEFAULT_FILES = ["stdout", "stderr", "mantik.log"]

    def __init__(self, files: t.Optional[t.List[str]] = None):
        self.directory = "."
        self._files = files or self.DEFAULT_FILES

    @property
    def properties(self):
        return {"foo": "bar"}

    def listdir(self, base="/"):
        return {
            f"{base}file1": FakeFileReference(content="File 1"),
            f"{base}file2": FakeFileReference(content="File 2"),
        }

    def stat(self, filename):
        if filename == "fake_dir":
            return FakeDirReference()
        if filename == "file_does_not_exist":
            raise FileNotFoundError()
        return FakeFileReference(content=f"Stat {filename}")


class FakeFileReference(pyunicore.client.PathFile):
    def __init__(self, content: str, name: str = "file"):
        self._content = content
        self.name = name

    def raw(self):
        return io.StringIO(self._content)

    def download(self, name: str) -> None:
        """we are not testing io, since unicore handels it"""
        pass


class FakeClient:
    def __init__(
        self,
        transport: FakeTransport = None,
        site_url: str = "test_api_url",
        login_successful: bool = False,
    ):
        if transport is None:
            transport = FakeTransport()
        self.transport = transport
        self.site_url = site_url
        self._properties = {"client": {"xlogin": {}}}
        if login_successful:
            self.add_login_info({"test_login": "test_logged_in"})

    @property
    def properties(self) -> t.Dict:
        return {**self.__dict__, **self._properties}

    def add_login_info(self, login: t.Dict) -> None:
        self._properties["client"]["xlogin"] = login

    def new_job(self, job_description: t.Dict, inputs: t.List) -> FakeJob:
        return FakeJob(transport=self.transport, job_url="test_job_url")

    def get_jobs(
        self, offset: int = 0, num: t.Optional[int] = None
    ) -> t.List[FakeJob]:
        total = (num or 10) + offset
        return [
            FakeJob(
                transport=self.transport,
                job_url=f"test_job_url{i}",
                submission_day=i + 1,
            )
            for i in range(total, offset - 1, -1)
        ]


class FakeDirReference(pyunicore.client.PathDir):
    def __init__(self, files: t.Optional[t.List[FakeFileReference]] = None):
        self.storage = FakeWorkDir(files or [])
        self.name = "fakedir"


def _create_run(
    properties: t.Optional[t.Dict] = None,
    status: _properties.Status = None,
    will_be_successful: bool = True,
) -> submitted_run.SubmittedRun:
    properties = properties or {}

    if status is not None:
        properties["status"] = status.value

    job = FakeJob(
        transport=FakeTransport(),
        job_url="test-job",
        properties=properties,
        will_be_successful=will_be_successful,
    )
    job = _job.Job(job)
    return submitted_run.SubmittedRun(run_id="test-job", job=job)



def test_import_fails_without_mlflow_installed():
    try:
        import mantik.mlflow  # noqa: F401
    except ModuleNotFoundError as e:
        if "Install supported MLflow version" not in str(e):
            assert (
                False
            ), "MLflow import should fail with dedicated error message"
        else:
            raise e



import enum
import typing as t

import firecrest

USER_NAME = "test-account"


class UploadStatus(enum.Enum):
    SUCCESS = "114"
    FAILED = "115"


class FakeParameter(t.TypedDict):
    name: str
    unit: str
    value: t.Any
    description: str


class FakeParameters(t.TypedDict):
    storage: t.List[FakeParameter]
    utilities: t.List[FakeParameter] = None


class FakeExternalUpload(firecrest.ExternalUpload):
    def __init__(self, status: UploadStatus):
        self._status = status
        self._in_progress = iter([True, False])

    def finish_upload(self) -> None:
        return None

    @property
    def in_progress(self) -> bool:
        return next(self._in_progress)

    @property
    def status(self) -> str:
        return self._status.value


class FakeJobSubmit(t.TypedDict):
    firecrest_taskid: str
    job_data_err: str
    job_data_out: str
    job_file: str
    job_file_err: str
    job_file_out: str
    jobid: int
    result: str


class FakeJobAcct(t.TypedDict):
    jobid: str
    name: str
    nodelist: str
    nodes: str
    partition: str
    start_time: str
    state: str
    time: str
    time_left: str
    user: str


class FakeClient(firecrest.Firecrest):
    def __init__(
        self,
        firecrest_url: str = "test-firecrest-url",
        authorization: t.Any = "test-authorization",
        verify: t.Union[str, bool, None] = None,
        sa_role: str = "firecrest-sa",
        job_status: str = "COMPLETED",
        parameters: t.Optional[FakeParameters] = None,
        upload_status: UploadStatus = UploadStatus.SUCCESS,
    ):
        self._firecrest_url = firecrest_url
        self._authorization = authorization
        self._verify = verify
        self._sa_role = sa_role
        self._job_status = job_status
        self._parameters = parameters
        self._upload_status = upload_status

    def mkdir(self, machine: str, target_path: str, p: bool = None) -> None:
        return None

    def parameters(self) -> FakeParameters:
        return self._parameters or FakeParameters(
            storage=[
                FakeParameter(
                    name="FILESYSTEMS",
                    unit="",
                    value=[
                        {
                            "mounted": [
                                "/scratch/snx3000",
                                "/project",
                                "/store",
                            ],
                            "system": "daint",
                        },
                        {
                            "mounted": ["/capstor/scratch/cscs"],
                            "system": "eiger",
                        },
                    ],
                    description="",
                ),
                FakeParameter(
                    name="OBJECT_STORAGE",
                    unit="",
                    value="swift",
                    description="",
                ),
                FakeParameter(
                    name="STORAGE_TEMPURL_EXP_TIME",
                    unit="seconds",
                    value="604800",
                    description="",
                ),
                FakeParameter(
                    name="STORAGE_MAX_FILE_SIZE",
                    unit="MB",
                    value="1024000",
                    description="",
                ),
            ],
            utilities=[
                FakeParameter(name="", unit="", value="", description="")
            ],
        )

    def whoami(self, machine=None) -> t.Optional[str]:
        return USER_NAME

    def external_upload(
        self, machine: str, source_path: str, target_path: str
    ) -> FakeExternalUpload:
        return FakeExternalUpload(status=self._upload_status)

    def submit(
        self,
        machine: str,
        job_script: t.Optional[str] = None,
        local_file: t.Optional[bool] = True,
        script_str: t.Optional[str] = None,
        script_local_path: t.Optional[str] = None,
        script_remote_path: t.Optional[str] = None,
        account: t.Optional[str] = None,
        env_vars: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> FakeJobSubmit:
        return FakeJobSubmit(
            firecrest_taskid="test-task-id-submit",
            job_data_err="",
            job_data_out="",
            job_file="",
            job_file_err="",
            job_file_out="",
            jobid="test-job-id",
            result="",
        )

    def poll(
        self,
        machine: str,
        jobs: t.Sequence[t.Union[str, int]] = None,
        start_time: str = None,
        end_time: str = None,
    ) -> t.List[FakeJobAcct]:
        return [
            FakeJobAcct(
                jobid="test-job-id",
                name="test-job",
                nodelist="None assigned",
                nodes="1",
                partition="test-partition",
                start_time="Unknown",
                state=self._job_status,
                time="Unknown",
                time_left="Unknown",
                user="test-user",
            )
        ]

    def poll_active(
        self, machine: str, jobs: t.Sequence[t.Union[str, int]] = None
    ) -> t.List:
        return []


class FakeClientCredentialsAuth:
    def __init__(
        self,
        client_id: str = "test-user",
        client_secret: str = "test-password",
        token_uri: str = "test-token-uri",
        login_successful: bool = True,
    ):
        self._client_id = client_id
        self._client_secret = client_secret
        self._token_uri = token_uri
        if login_successful:
            self.login_successful = True
        else:
            self.login_successful = False

    def get_access_token(self) -> str:
        if self.login_successful:
            return "test-token"
        else:
            raise firecrest.ClientsCredentialsException(
                responses=["test-response"]
            )



import os
import pathlib

import pytest

import mantik.config.environment as environment
import mantik.config.exceptions as exceptions
import mantik.config.executable as executable
import mantik.testing as testing


class TestEnvironment:
    @pytest.mark.parametrize(
        ("env_vars", "d", "expected"),
        [
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "Apptainer": {
                        "Path": "image.sif",
                    },
                },
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("image.sif")
                ),
            ),
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {"Python": "/path/to/venv"},
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
            ),
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {"Python": {"Path": "/path/to/venv"}},
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
            ),
            (
                testing.config.ALL_UNICORE_ENV_VARS,
                {
                    "Python": {"Path": "/path/to/venv"},
                    "PreRunCommandOnLoginNode": "precommand login node",
                    "PostRunCommandOnLoginNode": "postcommand login node",
                    "PreRunCommandOnComputeNode": "precommand compute node",
                    "PostRunCommandOnComputeNode": "postcommand compute node",
                },
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node=["precommand login node"],
                    post_run_command_on_login_node=["postcommand login node"],
                    pre_run_command_on_compute_node=["precommand compute node"],
                    post_run_command_on_compute_node=[
                        "postcommand compute node"
                    ],
                ),
            ),
        ],
    )
    def test_from_dict(
        self, monkeypatch, expect_raise_if_exception, env_vars, d, expected
    ):
        if "MLFLOW_TRACKING_TOKEN" in os.environ:
            del os.environ["MLFLOW_TRACKING_TOKEN"]
        if "MLFLOW_EXPERIMENT_ID" in os.environ:
            del os.environ["MLFLOW_EXPERIMENT_ID"]
        for key in env_vars:
            os.environ[key] = testing.config.DEFAULT_ENV_VAR_VALUE

        with expect_raise_if_exception(expected):
            result = environment.Environment.from_dict(d)

            assert result == expected

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: No execution environment
            (
                environment.Environment(),
                None,
            ),
            # Test case: Apptainer
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "srun apptainer run venv",
            ),
            # Test case: Python
            (
                testing.config._create_python_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                ),
                "source /path/to/venv/bin/activate",
            ),
        ],
    )
    def test_create_execution_command(self, env, expected):
        result = env._create_execution_command()

        assert expected == result

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: Apptainer with no modules and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=None, path=pathlib.Path("/path/to/venv")
                ),
                None,
            ),
            # Test case: Apptainer with no modules and precommand
            (
                testing.config._create_apptainer_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "a precommand\nanother precommand",
            ),
            # Test case: Apptainer with 1 module and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"], path=pathlib.Path("/path/to/venv")
                ),
                "module load Module",
            ),
            # Test case: Apptainer with 1 module and precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "module load Module\na precommand\nanother precommand",
            ),
            # Test case: Apptainer with 2 modules and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"], path=pathlib.Path("/path/to/venv")
                ),
                "module load Module",
            ),
            # Test case: Apptainer with 2 modules and precommand as a list
            (
                testing.config._create_apptainer_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "module load Module1 Module2\na precommand\n"
                "another precommand",
            ),
            # Test case: Python with no modules and no precommand
            (
                testing.config._create_python_environment(
                    modules=None, path=pathlib.Path("/path/to/venv")
                ),
                None,
            ),
            # Test case: Python with no modules and precommand
            (
                testing.config._create_python_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "a precommand\nanother precommand",
            ),
            # Test case: Python with 1 module and no precommand
            (
                testing.config._create_python_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                ),
                "module load Module",
            ),
            # Test case: Python with 1 module and precommand
            (
                testing.config._create_python_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "module load Module\na precommand\nanother precommand",
            ),
            # Test case: Python with 2 modules and no precommand
            (
                testing.config._create_python_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                ),
                "module load Module1 Module2",
            ),
            # Test case: Python with 2 modules and precommand
            (
                testing.config._create_python_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "module load Module1 Module2\na precommand\n"
                "another precommand",
            ),
            # Test case: No execution environment
            (
                environment.Environment(
                    modules=["Module1", "Module2"],
                    pre_run_command_on_compute_node=[
                        "a precommand",
                        "another precommand",
                    ],
                ),
                "module load Module1 Module2\na precommand\n"
                "another precommand",
            ),
        ],
    )
    def test_create_precommand_on_compute_node(self, env, expected):
        result = env._create_pre_run_command_on_compute_node()

        assert expected == result

    def test_create_postcommand_on_login_node(self):
        expected = "a postcommand && another postcommand"
        env = testing.config._create_python_environment(
            path=pathlib.Path("/path/to/venv"),
            post_run_command_on_login_node=[
                "a postcommand",
                "another postcommand",
            ],
        )

        result = env._create_login_node_command(
            env.post_run_command_on_login_node
        )

        assert expected == result

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: No execution environment
            (
                environment.Environment(
                    modules=["Module1", "Module2"],
                    pre_run_command_on_compute_node=["a precommand"],
                    post_run_command_on_compute_node=["a postcommand"],
                ),
                (
                    "module load Module1 Module2\n"
                    "a precommand && "
                    "python main.py test && "
                    "a postcommand"
                ),
            ),
            # Test case: Apptainer with no precommand, no postcommand, no modules # noqa E501
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "srun apptainer run venv python main.py test",
            ),
            # Test case: Apptainer with a precommand, a postcommand and modules
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv"),
                    modules=["Module1", "Module2"],
                    pre_run_command_on_compute_node=["a precommand"],
                    post_run_command_on_compute_node=["a postcommand"],
                ),
                (
                    "module load Module1 Module2\n"
                    "a precommand && "
                    "srun apptainer run venv "
                    "python main.py test && "
                    "a postcommand"
                ),
            ),
            # Test case: Python with no precommand, no postcommand, no modules
            (
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "source /path/to/venv/bin/activate && python main.py test",
            ),
            # Test case: Python with a precommand, a postcommand and modules
            (
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv"),
                    modules=["Module1", "Module2"],
                    pre_run_command_on_compute_node=["a precommand"],
                    post_run_command_on_compute_node=["a postcommand"],
                ),
                (
                    "module load Module1 Module2\n"
                    "a precommand && "
                    "source /path/to/venv/bin/activate && "
                    "python main.py test && "
                    "a postcommand"
                ),
            ),
        ],
    )
    def test_create_execution_command_with_arguments(self, env, expected):
        entry_point_arguments = "python main.py test"
        result = env._create_execution_command_with_arguments(
            entry_point_arguments
        )
        assert result == expected

    @pytest.mark.parametrize(
        "env",
        [
            # Test case: Apptainer with no variables
            testing.config._create_apptainer_environment(
                modules=None,
                path=pathlib.Path("/path/to/venv"),
                include_mlflow_env_vars=False,
            ),
            # Test case: Python with no variables
            testing.config._create_python_environment(
                modules=None,
                path=pathlib.Path("/path/to/venv"),
                include_mlflow_env_vars=False,
            ),
        ],
    )
    def test_add_env_variables(self, env):
        assert env.variables is None
        expected = {"env_var": "ENV_VAR"}
        env.add_env_vars(expected)
        result = env.variables
        assert expected == result

    @pytest.mark.parametrize(
        ("config", "expected", "error_message"),
        [
            # Test case: No execution environment given
            (
                {"Resources": {}},
                None,
                None,
            ),
            # Test case: One execution environment given
            (
                {"Apptainer": {"Path": "image.sif"}},
                executable.Apptainer(path=pathlib.Path("image.sif")),
                None,
            ),
            # Test case: Two execution environments given, should raise
            (
                {"Apptainer": {}, "Python": {}},
                exceptions.ConfigValidationError(),
                "Only one execution environment is allowed, "
                "but in config these have been "
                "found: 'Apptainer', 'Python'.",
            ),
        ],
    )
    def test_get_execution_environment(
        self, expect_raise_if_exception, config, expected, error_message
    ):
        with expect_raise_if_exception(expected) as e:
            result = environment._get_execution_environment(config)
            assert result == expected

        if error_message is not None:
            assert str(e.value) == error_message

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: Apptainer with no modules and no env vars
            (
                testing.config._create_apptainer_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    include_mlflow_env_vars=False,
                ),
                (
                    "export MANTIK_WORKING_DIRECTORY=/test-scratch-dir/test-user-name/mantik/test-run-dir\n"  # noqa E501
                    "srun apptainer run venv python main.py test"
                ),
            ),
            # Test case: Apptainer with 1 module and env vars
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    variables={"TEST_ENV_VAR": "test-value"},
                    include_mlflow_env_vars=False,
                ),
                (
                    "export MANTIK_WORKING_DIRECTORY=/test-scratch-dir/test-user-name/mantik/test-run-dir\n"  # noqa E501
                    "export TEST_ENV_VAR=test-value\n"
                    "module load Module && "
                    "srun apptainer run venv python main.py test"
                ),
            ),
            # Test case: Python with no modules and no env vars
            (
                testing.config._create_python_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    include_mlflow_env_vars=False,
                ),
                (
                    "export MANTIK_WORKING_DIRECTORY=/test-scratch-dir/test-user-name/mantik/test-run-dir\n"  # noqa E501
                    "source /path/to/venv/bin/activate && python main.py test"
                ),
            ),
            # Test case: Python with 1 module and env vars
            (
                testing.config._create_python_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    variables={"TEST_ENV_VAR": "test-value"},
                    include_mlflow_env_vars=False,
                ),
                (
                    "export MANTIK_WORKING_DIRECTORY=/test-scratch-dir/test-user-name/mantik/test-run-dir\n"  # noqa E501
                    "export TEST_ENV_VAR=test-value\n"
                    "module load Module && "
                    "source /path/to/venv/bin/activate && "
                    "python main.py test"
                ),
            ),
            # Test case: No execution environment
            (
                environment.Environment(
                    modules=["Module1", "Module2"],
                    variables={"TEST_ENV_VAR": "test-value"},
                ),
                (
                    "export MANTIK_WORKING_DIRECTORY=/test-scratch-dir/test-user-name/mantik/test-run-dir\n"  # noqa E501
                    "export TEST_ENV_VAR=test-value\n"
                    "module load Module1 Module2 && "
                    "python main.py test"
                ),
            ),
        ],
    )
    def test_to_slurm_batch_script(self, env, expected):
        entry_point_arguments = "python main.py test"
        result = env.to_slurm_batch_script(
            entry_point_arguments=entry_point_arguments,
            run_dir=pathlib.Path(
                "/test-scratch-dir/test-user-name/mantik/test-run-dir"
            ),
        )

        assert result == expected



import dataclasses
import typing as t
import uuid

import mantik.utils.mantik_api.client as client
import mantik.utils.mantik_api.user as _user


@dataclasses.dataclass(frozen=True)
class Connection:
    """
    Model for the Connection credentials delivered by Mantik API.

    Attributes
    ----------
    name: the name of the Connection
    username: the username of the Connection credentials
    password: the password of the Connection credentials
    """

    connection_id: uuid.UUID
    user: _user.User
    connection_name: str
    connection_provider: str
    auth_method: str
    login_name: t.Optional[str] = None
    password: t.Optional[str] = None
    token: t.Optional[str] = None

    @classmethod
    def from_json(cls, data: t.Dict) -> "Connection":
        return cls(
            connection_id=uuid.UUID(data["connectionId"]),
            user=_user.User.from_json(data["user"]),
            connection_name=data["connectionName"],
            connection_provider=data["connectionProvider"],
            auth_method=data["authMethod"],
            login_name=data.get("loginName"),
            password=data.get("password"),
            token=data.get("token"),
        )


def get(
    user_id: uuid.UUID,
    connection_id: uuid.UUID,
    token: str,
) -> Connection:
    endpoint = f"/users/{user_id}/settings/connections/{connection_id}"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    ).json()

    return Connection.from_json(response)



import functools
import json
import pathlib
import uuid

import fastapi.testclient
import mlflow.projects
import pytest

import mantik.config as config
import mantik.testing as testing
import mantik.utils.unicore.zip as unicore_zip
import mantik_compute_backend
import mantik_compute_backend.settings as settings
import tokens.verifier as _verifier


@pytest.fixture(scope="function")
def client(monkeypatch) -> fastapi.testclient.TestClient:
    monkeypatch.setattr(
        _verifier,
        "TokenVerifier",
        testing.mlflow_server.FakeTokenVerifier,
    )
    app = mantik_compute_backend.app.create_app()

    return fastapi.testclient.TestClient(app)


@pytest.fixture(scope="function")
def client_suppressing_raise(monkeypatch) -> fastapi.testclient.TestClient:
    monkeypatch.setattr(
        _verifier,
        "TokenVerifier",
        testing.mlflow_server.FakeTokenVerifier,
    )
    app = mantik_compute_backend.app.create_app()

    return fastapi.testclient.TestClient(app, raise_server_exceptions=False)


@pytest.fixture(scope="function")
def client_with_small_size_limitation(
    monkeypatch,
) -> fastapi.testclient.TestClient:
    monkeypatch.setattr(
        _verifier,
        "TokenVerifier",
        testing.mlflow_server.FakeTokenVerifier,
    )
    app = mantik_compute_backend.app.create_app()

    def get_settings_override() -> settings.Settings:
        return settings.Settings(max_file_size=1)

    app.dependency_overrides[settings.get_settings] = get_settings_override

    return fastapi.testclient.TestClient(app, raise_server_exceptions=False)


@pytest.fixture
def submit_run_request_data():
    return {
        "active_run_id": str(uuid.uuid4().hex),
        "entry_point": "main",
        "mlflow_parameters": json.dumps({"foo": "bar"}),
        "hpc_api_username": "bar",
        "hpc_api_password": "bam",
        "compute_budget_account": "baz",
        "compute_backend_config": "compute-backend-config.json",
        "experiment_id": "1",
        "mlflow_tracking_uri": "foo",
        "mlflow_tracking_token": "aasdf",
    }


@pytest.fixture(scope="session")
def mlproject_path() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent / "../../../tests/resources/test-project"
    )


@pytest.fixture()
def example_project(
    mlproject_path,
) -> mlflow.projects._project_spec.Project:
    return mlflow.projects.utils.load_project(mlproject_path)


@pytest.fixture(scope="function")
def example_unicore_config() -> config.core.Config:
    return config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            ),
            variables={"SRUN_CPUS_PER_TASK": 100},
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )


@pytest.fixture()
def example_unicore_config_for_python() -> config.core.Config:
    return config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Python(
                path=pathlib.Path("/venv"),
            ),
            pre_run_command_on_compute_node=["precommand compute node"],
            post_run_command_on_compute_node=["postcommand compute node"],
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )


@pytest.fixture(scope="function")
def example_firecrest_config() -> config.core.Config:
    return config.core.Config(
        firecrest=config.firecrest.Firecrest(
            api_url="test-url",
            machine="test-machine",
            token_url="test-token-url",
        ),
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            )
        ),
        resources=config.resources.Resources(
            queue="batch", node_constraints="gpu"
        ),
        exclude=[],
    )


@pytest.fixture(scope="function")
def example_firecrest_config_long_resources() -> config.core.Config:
    return config.core.Config(
        firecrest=config.firecrest.Firecrest(
            api_url="test-url",
            machine="test-machine",
            token_url="test-token-url",
        ),
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            )
        ),
        resources=config.resources.Resources(
            queue="batch",
            node_constraints="gpu",
            runtime="1h",
            nodes=1,
            total_cpus=1,
            cpus_per_node=1,
            gpus_per_node=1,
            memory_per_node="1",
            reservation="test-reservation",
            qos="test-qos,",
        ),
        exclude=[],
    )


@pytest.fixture()
def example_firecrest_config_for_python() -> config.core.Config:
    return config.core.Config(
        firecrest=config.firecrest.Firecrest(
            api_url="test-url",
            machine="test-machine",
            token_url="test-token-url",
        ),
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Python(
                path=pathlib.Path("/venv"),
            ),
            pre_run_command_on_compute_node=["precommand compute node"],
            post_run_command_on_compute_node=["postcommand compute node"],
        ),
        resources=config.resources.Resources(
            queue="batch", node_constraints="gpu"
        ),
        exclude=["*.sif"],
    )


@pytest.fixture()
def zipped_unicore_content(mlproject_path, example_unicore_config):
    return unicore_zip.zip_directory_with_exclusion(
        mlproject_path, example_unicore_config
    )


@pytest.fixture()
def zipped_firecrest_content(mlproject_path, example_firecrest_config):
    return unicore_zip.zip_directory_with_exclusion(
        mlproject_path, example_firecrest_config
    )


@pytest.fixture()
def submit_run_request_files(zipped_unicore_content):
    return {"mlproject_zip": zipped_unicore_content}


@pytest.fixture(scope="session")
def broken_mlproject_path() -> pathlib.Path:
    return pathlib.Path(__file__).parent / "../../resources/broken-project"


@pytest.fixture(scope="function")
def broken_config() -> config.core.Config:
    return config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("/mantik-test.sif"), type="remote"
            )
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=[],
    )


@pytest.fixture()
def broken_zipped_content(broken_mlproject_path, broken_config):
    return unicore_zip.zip_directory_with_exclusion(
        broken_mlproject_path, broken_config
    )


@pytest.fixture()
def example_broken_project(
    broken_mlproject_path,
) -> mlflow.projects._project_spec.Project:
    return mlflow.projects.utils.load_project(broken_mlproject_path)


@pytest.fixture()
def create_firecrest_fake_client_credentials_auth(
    login_successful: bool,
) -> functools.partial:
    return functools.partial(
        testing.firecrest.FakeClientCredentialsAuth,
        login_successful=login_successful,
    )



import functools

import pytest
import pyunicore.client as pyunicore

import mantik.testing as testing
import mantik_compute_backend.unicore as unicore
import mantik_compute_backend.unicore.connect as _connect


@pytest.fixture()
def transport():
    return testing.pyunicore.FakeTransport()


def create_fake_client(login_successful: bool) -> functools.partial:
    return functools.partial(
        testing.pyunicore.FakeClient,
        login_successful=login_successful,
    )


@pytest.mark.parametrize(
    ("login_successful", "expected"),
    [
        (False, unicore.exceptions.AuthenticationFailedException()),
        (True, testing.pyunicore.FakeClient),
    ],
)
def test_create_unicore_api_connection(
    expect_raise_if_exception, monkeypatch, login_successful, expected
):
    monkeypatch.setattr(pyunicore, "Transport", testing.pyunicore.FakeTransport)
    monkeypatch.setattr(
        pyunicore,
        "Client",
        create_fake_client(login_successful=login_successful),
    )

    api_url = "test-api-url"
    user = "test_user"
    password = "test_password"

    with expect_raise_if_exception(expected):
        result = _connect.create_unicore_api_connection(
            api_url=api_url,
            user=user,
            password=password,
        )

        assert isinstance(result, expected)


def test_create_token():
    user = "test_user"
    password = "test_password"
    expected = "dGVzdF91c2VyOnRlc3RfcGFzc3dvcmQ="

    result = _connect._create_token(user=user, password=password)

    assert result == expected


@pytest.mark.parametrize(
    ("login", "expected"),
    [
        ({}, True),
        ({"test_login_info": "test_login"}, False),
    ],
)
def test_authentication_failed(login, expected):
    client = testing.pyunicore.FakeClient()
    client.add_login_info(login)

    result = _connect._authentication_failed(client)

    assert result == expected



import contextlib
import datetime
import os
import typing as t

import pytest

import mantik.authentication as authentication
import mantik.utils as utils
import mantik.utils.mantik_api.client as mantik_api
import mantik.utils.mantik_api.credentials as _credentials

EXPIRES_AT = datetime.datetime(2022, 1, 1)


@pytest.fixture()
def credentials() -> _credentials.Credentials:
    return _credentials.Credentials(
        username="test-user",
        password="test-password",
    )


def test_create_tokens(requests_mock, credentials):
    requests_mock.post(
        url=f"{mantik_api._DEFAULT_MANTIK_API_URL}"
        f"{mantik_api.MANTIK_API_CREATE_TOKEN_API_PATH}",
        json={
            "AccessToken": "test-access-token",
            "RefreshToken": "test-refresh-token",
            "ExpiresAt": EXPIRES_AT.isoformat(),
        },
    )

    result = authentication.api.create_tokens(credentials=credentials)

    assert result == authentication.tokens.Tokens(
        access_token="test-access-token",
        refresh_token="test-refresh-token",
        expires_at=EXPIRES_AT,
    )


def test_refresh_tokens(
    requests_mock,
    credentials,
):
    requests_mock.post(
        url=f"{mantik_api._DEFAULT_MANTIK_API_URL}"
        f"{mantik_api.MANTIK_API_REFRESH_TOKEN_API_PATH}",
        json={
            "AccessToken": "test-refreshed-access-token",
            "RefreshToken": "test-refreshed-refresh-token",
            "ExpiresAt": EXPIRES_AT.isoformat(),
        },
    )

    tokens = authentication.tokens.Tokens(
        access_token="test-access-token",
        refresh_token="test-refresh-token",
        expires_at=datetime.datetime(2000, 1, 1),
    )

    result = authentication.api.refresh_tokens(
        credentials=credentials,
        tokens=tokens,
    )

    assert result == authentication.tokens.Tokens(
        access_token="test-refreshed-access-token",
        refresh_token="test-refresh-token",
        expires_at=EXPIRES_AT,
    )


@contextlib.contextmanager
def _set_mlflow_tracking_uri_env_var(value: t.Optional[str]) -> None:
    if os.environ.get(utils.mlflow.TRACKING_URI_ENV_VAR):
        os.environ.pop(utils.mlflow.TRACKING_URI_ENV_VAR)
    if value is not None:
        os.environ[utils.mlflow.TRACKING_URI_ENV_VAR] = value
    yield
    if os.environ.get(utils.mlflow.TRACKING_URI_ENV_VAR):
        os.environ.pop(utils.mlflow.TRACKING_URI_ENV_VAR)



import dataclasses
import itertools
import logging
import time

import mlflow.entities
import pandas as pd

import mantik
import mantik.testing.stress as stress

logger = logging.getLogger()
stress.setup_logger(logger)


@dataclasses.dataclass
class Results(stress.ResultsBase):
    n_params: int
    duration_per_param: float = None

    def __post_init__(self):
        self.duration_per_param = self.duration / self.n_params
        logger.info(
            (
                "Sending %s params via mlflow.%s %s %s seconds, "
                "average duration per param was %s seconds"
            ),
            self.n_params,
            self.mlflow_method,
            "failed after" if self.failed else "succeeded in",
            self.duration,
            self.duration_per_param,
        )

    def to_table_row(self) -> str:
        return (
            f"| {self.n_params} "
            f"| {self.mlflow_method} "
            f"| {self.duration} "
            f"| {self.duration_per_param} "
            f"| {not self.failed} "
            f"|\n"
        )

    def insert_to_df(self, df: pd.DataFrame) -> None:
        row = df.loc[self.n_params, self.mlflow_method]
        row["duration"] = self.duration
        row["duration per param"] = self.duration_per_param
        row["succeeded"] = not self.failed


def run_stress_test(sleep_time: int, dry_run: bool = False):
    results = []

    values = [10, 100, 1000, 10000]
    methods = ["log_param", "log_params", "log_batch"]
    index = pd.MultiIndex.from_arrays(
        list(zip(*itertools.product(values, methods))),
        names=["# params", "method"],
    )

    df = pd.DataFrame(
        data=None,
        index=index,
        columns=["duration", "duration per param", "succeeded"],
    )

    context = stress.get_context(dry_run)

    for n_params in values:
        for method in [
            _log_individual_params_for_run,
            _log_grouped_params_for_run,
            _log_batched_params_for_run,
        ]:
            logger.info(
                "Running test for logging %s params via %s",
                n_params,
                method.__name__,
            )
            result = method(n_params=n_params, dry_run=dry_run, context=context)
            result.insert_to_df(df)
            results.append(result)
            logger.info(
                "Sleeping for %s seconds until next stress test", sleep_time
            )
            time.sleep(sleep_time)
    df.to_csv("results-parameters.csv")
    logger.info("Results saved to results-parameters.csv")
    stress.log_results(
        results,
        columns=[
            "# params",
            "method",
            "duration",
            "duration per param",
            "succeeded",
        ],
        logger=logger,
    )


def _log_individual_params_for_run(
    n_params: int,
    dry_run: bool,
    context: stress.Context,
) -> Results:
    failed = False
    start = time.time()
    with context(run_name=f"test-requests-individual-{n_params}"):
        for i in range(n_params):
            try:
                if not dry_run:
                    mlflow.log_param(key=f"param-{i}", value=1)
            except:  # noqa: E722
                logger.exception(
                    "Exception after %s individual params", i, exc_info=True
                )
                failed = True
                break
    duration = time.time() - start
    return Results(
        n_params=n_params,
        mlflow_method="log_param",
        duration=duration,
        failed=failed,
    )


def _log_grouped_params_for_run(
    n_params: int,
    dry_run: bool,
    context: stress.Context,
) -> Results:
    failed = False
    params = {f"param-{i}": 1 for i in range(n_params)}
    start = time.time()
    with context(run_name=f"test-requests-grouped-{n_params}"):
        try:
            if not dry_run:
                mlflow.log_params(params)
        except:  # noqa: E722
            logger.exception(
                "Exception sending %s grouped params", n_params, exc_info=True
            )
            failed = True
    duration = time.time() - start
    return Results(
        n_params=n_params,
        mlflow_method="log_params",
        duration=duration,
        failed=failed,
    )


def _log_batched_params_for_run(
    n_params: int, dry_run: bool, context: stress.Context
) -> Results:
    failed = False
    params = [
        mlflow.entities.Param(key=f"param-{i}", value=str(1))
        for i in range(n_params)
    ]
    start = time.time()
    with context(run_name=f"test-requests-grouped-{n_params}") as run:
        client = mlflow.MlflowClient()
        try:
            if not dry_run:
                client.log_batch(run_id=run.info.run_id, params=params)
        except:  # noqa: E722
            logger.exception(
                "Exception sending %s batched params", n_params, exc_info=True
            )
            failed = True
    duration = time.time() - start
    return Results(
        n_params=n_params,
        mlflow_method="log_batch",
        duration=duration,
        failed=failed,
    )


if __name__ == "__main__":
    mantik.init_tracking()
    args = stress.setup_argpaser().parse_args()
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    run_stress_test(sleep_time=int(args.sleep_time), dry_run=args.dry_run)



import dataclasses
import typing as t

import mantik.utils as utils


@dataclasses.dataclass
class Environment:
    """The required environment variables for tracking."""

    token: str

    def to_dict(self) -> t.Dict[str, str]:
        """Return as dict."""
        return {
            utils.mlflow.TRACKING_TOKEN_ENV_VAR: self.token,
        }

    def to_bash_statement(self, no_export: bool = False) -> str:
        """Return as a bash statement.

        Parameters
        ----------
        no_export : bool, default=False
            Whether to include the `export` command as a prefix.

        Returns
        -------
        str
            The environment variables as a bash statement.
            E.g. `export ENV_VAR=<value>`.

        """
        statement = f"{utils.mlflow.TRACKING_TOKEN_ENV_VAR}={self.token}"
        if no_export:
            return statement
        return f"export {statement}"



import dataclasses
import logging
import typing as t
import uuid

import mantik.utils.mantik_api.client as client
import mantik.utils.other as other_utils

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class GPUInfo:
    name: str
    id: str
    driver: str
    total_memory: str

    @classmethod
    def from_dict(cls, _dict):
        return cls(
            name=_dict["name"],
            id=_dict["id"],
            driver=_dict["driver"],
            total_memory=_dict["totalMemory"],
        )

    def to_json(self):
        return {
            "name": self.name,
            "id": self.id,
            "driver": self.driver,
            "totalMemory": self.total_memory,
        }


@dataclasses.dataclass
class RunInfrastructure:
    os: str
    cpu_cores: int
    gpu_count: int
    gpu_info: t.List[GPUInfo]
    hostname: str
    memory_gb: str
    platform: str
    processor: str
    python_version: str
    python_executable: str

    def to_json(self):
        return {
            "os": self.os,
            "cpuCores": self.cpu_cores,
            "gpuCount": self.gpu_count,
            "gpuInfo": [_gpu.to_json() for _gpu in self.gpu_info],
            "hostname": self.hostname,
            "memoryGb": self.memory_gb,
            "platform": self.platform,
            "processor": self.processor,
            "pythonVersion": self.python_version,
            "pythonExecutable": self.python_executable,
        }

    @classmethod
    def from_system(cls):
        _gpu_info = other_utils.get_gpu_info()
        return cls(
            os=other_utils.get_os(),
            cpu_cores=other_utils.get_cpu_cores(),
            gpu_count=len(_gpu_info),
            gpu_info=[GPUInfo.from_dict(_gpu) for _gpu in _gpu_info],
            hostname=other_utils.get_hostname(),
            memory_gb=other_utils.get_memory_gb(),
            platform=other_utils.get_platform(),
            processor=other_utils.get_processor(),
            python_version=other_utils.get_python_version(),
            python_executable=other_utils.get_python_executable(),
        )


def submit_run(project_id: uuid.UUID, submit_run_data: dict, token: str):
    endpoint = f"/projects/{project_id}/runs"
    response = client.send_request_to_mantik_api(
        method="POST", data=submit_run_data, url_endpoint=endpoint, token=token
    )
    logger.info("Run has been successfully submitted")
    return response


def save_run(project_id: uuid.UUID, run_data: dict, token: str):
    endpoint = f"/projects/{project_id}/runs"
    response = client.send_request_to_mantik_api(
        method="POST",
        data=run_data,
        url_endpoint=endpoint,
        token=token,
        query_params={"submit": False},
    )
    logger.info("Run has been successfully saved")
    return response


def update_run_status(
    project_id: uuid.UUID, run_id: uuid.UUID, status: str, token: str
):
    endpoint = f"/projects/{project_id}/runs/{run_id}/status"
    response = client.send_request_to_mantik_api(
        method="PUT", data=status, url_endpoint=endpoint, token=token
    )
    logger.info("Run status has been successfully updated")
    return response


def update_logs(
    project_id: uuid.UUID, run_id: uuid.UUID, logs: str, token: str
):
    endpoint = f"/projects/{project_id}/runs/{run_id}/logs"
    response = client.send_request_to_mantik_api(
        method="PUT", data=logs, url_endpoint=endpoint, token=token
    )
    logger.info("Run logs has been successfully updated")
    return response


def get_download_artifact_url(
    project_id: uuid.UUID, run_id: uuid.UUID, token: str
):
    endpoint = f"/projects/{project_id}/runs/{run_id}/artifacts"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    logger.info("Artifacts' download url successfully fetched")
    return response.json()["url"]


def update_run_infrastructure(
    project_id: uuid.UUID,
    run_id: uuid.UUID,
    infrastructure: RunInfrastructure,
    token: str,
):
    endpoint = f"/projects/{project_id}/runs/{run_id}/infrastructure"
    response = client.send_request_to_mantik_api(
        method="PUT",
        data=infrastructure.to_json(),
        url_endpoint=endpoint,
        token=token,
    )
    logger.info("Run infrastructure has been successfully updated")
    return response
